#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot/blob/master/LICENSE

from __future__ import absolute_import

import copy
import re
import ast
from functools import reduce

import uproot.const
import uproot.rootio
from uproot.interp.numerical import asdtype
from uproot.interp.numerical import asarray
from uproot.interp.numerical import asdouble32
from uproot.interp.numerical import asfloat16
from uproot.interp.numerical import asstlbitset
from uproot.interp.jagged import asjagged
from uproot.interp.objects import astable
from uproot.interp.objects import asobj
from uproot.interp.objects import asgenobj
from uproot.interp.objects import asstring
from uproot.interp.objects import SimpleArray
from uproot.interp.objects import STLVector
from uproot.interp.objects import STLMap
from uproot.interp.objects import STLString
from uproot.interp.objects import Pointer

class _NotNumerical(Exception): pass

def _normalize_ftype(fType):
    if fType is not None and uproot.const.kOffsetL < fType < uproot.const.kOffsetP:
        return fType - uproot.const.kOffsetL
    else:
        return fType

def _ftype2dtype(fType, awkward):
    fType = _normalize_ftype(fType)
    if fType == uproot.const.kBool:
        return awkward.numpy.dtype(awkward.numpy.bool_)
    elif fType == uproot.const.kChar:
        return awkward.numpy.dtype("i1")
    elif fType == uproot.const.kUChar:
        return awkward.numpy.dtype("u1")
    elif fType == uproot.const.kShort:
        return awkward.numpy.dtype(">i2")
    elif fType == uproot.const.kUShort:
        return awkward.numpy.dtype(">u2")
    elif fType == uproot.const.kInt:
        return awkward.numpy.dtype(">i4")
    elif fType in (uproot.const.kBits, uproot.const.kUInt, uproot.const.kCounter):
        return awkward.numpy.dtype(">u4")
    elif fType == uproot.const.kLong:
        return awkward.numpy.dtype(">i8")
    elif fType == uproot.const.kULong:
        return awkward.numpy.dtype(">u8")
    elif fType == uproot.const.kLong64:
        return awkward.numpy.dtype(">i8")
    elif fType == uproot.const.kULong64:
        return awkward.numpy.dtype(">u8")
    elif fType == uproot.const.kFloat:
        return awkward.numpy.dtype(">f4")
    elif fType == uproot.const.kDouble:
        return awkward.numpy.dtype(">f8")
    else:
        raise _NotNumerical

def _leaf2dtype(leaf, awkward):
    classname = leaf.__class__.__name__
    if classname == "TLeafO":
        return awkward.numpy.dtype(awkward.numpy.bool_)
    elif classname == "TLeafB":
        if leaf._fIsUnsigned:
            return awkward.numpy.dtype(awkward.numpy.uint8)
        else:
            return awkward.numpy.dtype(awkward.numpy.int8)
    elif classname == "TLeafS":
        if leaf._fIsUnsigned:
            return awkward.numpy.dtype(awkward.numpy.uint16)
        else:
            return awkward.numpy.dtype(awkward.numpy.int16)
    elif classname == "TLeafI":
        if leaf._fIsUnsigned:
            return awkward.numpy.dtype(awkward.numpy.uint32)
        else:
            return awkward.numpy.dtype(awkward.numpy.int32)
    elif classname == "TLeafL":
        if leaf._fIsUnsigned:
            return awkward.numpy.dtype(awkward.numpy.uint64)
        else:
            return awkward.numpy.dtype(awkward.numpy.int64)
    elif classname == "TLeafF":
        return awkward.numpy.dtype(awkward.numpy.float32)
    elif classname == "TLeafD":
        return awkward.numpy.dtype(awkward.numpy.float64)
    elif classname == "TLeafElement":
        return _ftype2dtype(leaf._fType, awkward)
    else:
        raise _NotNumerical

def _obj_or_genobj(streamerClass, branch, isjagged, cntvers=False, tobject=True, speedbump=True):
    if len(branch._fBranches) != 0:
        return None

    try:
        recarray = streamerClass._recarray_dtype(cntvers=cntvers, tobject=tobject)

    except (AttributeError, ValueError):
        if not speedbump:
            context = copy.copy(branch._context)
            context.speedbump = False
        else:
            context = branch._context

        if isjagged:
            return asgenobj(SimpleArray(streamerClass), context, 0)
        else:
            return asgenobj(streamerClass, context, 0)

    else:
        if isjagged:
            if streamerClass._methods is None:
                return asjagged(astable(asdtype(recarray)))
            else:
                return asjagged(asobj(astable(asdtype(recarray)), streamerClass._methods))
        else:
            if streamerClass._methods is None:
                return asdtype(recarray)
            else:
                return asobj(astable(asdtype(recarray)), streamerClass._methods)

def interpret(branch, awkwardlib=None, swapbytes=True, cntvers=False, tobject=True, speedbump=True):
    import uproot.tree
    awkward = uproot.tree._normalize_awkwardlib(awkwardlib)

    dims, isjagged = (), False
    if len(branch._fLeaves) == 1:
        m = interpret._titlehasdims.match(branch._fLeaves[0]._fTitle)
        if m is not None:
            dims = tuple(int(x) for x in re.findall(interpret._itemdimpattern, branch._fLeaves[0]._fTitle))
            if dims == ():
                dims = (branch._fLeaves[0]._fLen, ) if branch._fLeaves[0]._fLen > 1 else ()
            if any(interpret._itemdimpattern.match(x) is None for x in re.findall(interpret._itemanypattern, branch._fLeaves[0]._fTitle)):
                isjagged = True
    else:
        for leaf in branch._fLeaves:
            if interpret._titlehasdims.match(leaf._fTitle):
                return None

    try:
        if len(branch._fLeaves) == 1:
            if isinstance(branch._streamer, uproot.rootio.TStreamerObjectPointer):
                obj = branch._streamer._fTypeName
                if obj.endswith(b"*"):
                    obj = obj[:-1]
                obj = uproot.rootio._safename(obj)
                if obj in branch._context.classes:
                    return _obj_or_genobj(branch._context.classes.get(obj), branch, isjagged, cntvers=cntvers, tobject=tobject, speedbump=speedbump)

            # Process Double32_t and Float16_t types possibly packed in TLeafElement
            leaftype = uproot.const.kBase
            if branch._fLeaves[0].__class__.__name__ == "TLeafElement":
                leaftype = _normalize_ftype(branch._fLeaves[0]._fType)

            iskDouble32 = leaftype == uproot.const.kDouble32
            iskFloat16  = leaftype == uproot.const.kFloat16

            if iskDouble32 or iskFloat16:
                def transform(node, tofloat=True):
                    if isinstance(node, ast.AST):
                        if isinstance(node, ast.Name) and isinstance(node.ctx, ast.Load) and node.id == "pi":
                            out = ast.Num(3.141592653589793)  # TMath::Pi()
                        elif isinstance(node, ast.Num):
                            out = ast.Num(float(node.n))
                        elif isinstance(node, ast.BinOp) and isinstance(node.op, (ast.Add, ast.Sub, ast.Mult, ast.Div)):
                            out = ast.BinOp(transform(node.left), node.op, transform(node.right))
                        elif isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.USub):
                            out = ast.UnaryOp(node.op, transform(node.operand))
                        elif isinstance(node, ast.List) and isinstance(node.ctx, ast.Load) and len(node.elts) == 2:
                            out = ast.List([transform(node.elts[0]), transform(node.elts[1])], node.ctx)
                        elif isinstance(node, ast.List) and isinstance(node.ctx, ast.Load) and len(node.elts) == 3 and isinstance(node.elts[2], ast.Num):
                            out = ast.List([transform(node.elts[0]), transform(node.elts[1]), node.elts[2]], node.ctx)
                        else:
                            raise Exception(ast.dump(node))
                        out.lineno, out.col_offset = node.lineno, node.col_offset
                        return out
                    else:
                        raise Exception(ast.dump(node))

                try:
                    left, right = branch._streamer._fTitle.index(b"["), branch._streamer._fTitle.index(b"]")
                except (ValueError, AttributeError):
                    low, high, numbits = 0, 0, 0
                else:
                    try:
                        spec = eval(compile(ast.Expression(transform(ast.parse(branch._streamer._fTitle[left : right + 1]).body[0].value)), repr(branch._streamer._fTitle), "eval"))
                        if len(spec) == 2:
                            low, high = spec
                            numbits = None
                        else:
                            low, high, numbits = spec
                    except Exception:
                        return None

                if iskDouble32 and numbits == 0:
                    out = asdtype(awkward.numpy.dtype((">f4", dims)), awkward.numpy.dtype(("f8", dims)))
                elif iskDouble32 and numbits is None:
                    out = asdouble32(low, high, 32, dims)
                elif iskDouble32:
                    out = asdouble32(low, high, numbits, dims)
                elif iskFloat16 and numbits == 0:
                    out = asfloat16(low, high, 12, dims)
                elif iskFloat16 and numbits is None:
                    out = asfloat16(low, high, 32, dims)
                elif iskFloat16:
                    out = asfloat16(low, high, numbits, dims)
                else:
                    return None

            else:
                fromdtype = _leaf2dtype(branch._fLeaves[0], awkward).newbyteorder(">")

                if swapbytes:
                    out = asdtype(awkward.numpy.dtype((fromdtype, dims)), awkward.numpy.dtype((fromdtype.newbyteorder("="), dims)))
                else:
                    out = asdtype(awkward.numpy.dtype((fromdtype, dims)), awkward.numpy.dtype((fromdtype, dims)))

            if branch._fLeaves[0]._fLeafCount is None:
                return out
            else:
                return asjagged(out)

        elif len(branch._fLeaves) > 1:
            fromdtype = awkward.numpy.dtype([(str(leaf._fName.decode("ascii")), _leaf2dtype(leaf, awkward).newbyteorder(">")) for leaf in branch._fLeaves])
            if swapbytes:
                todtype = awkward.numpy.dtype([(str(leaf._fName.decode("ascii")), _leaf2dtype(leaf, awkward).newbyteorder("=")) for leaf in branch._fLeaves])
            else:
                todtype = fromdtype

            if all(leaf._fLeafCount is None for leaf in branch._fLeaves):
                return asdtype(awkward.numpy.dtype((fromdtype, dims)), awkward.numpy.dtype((todtype, dims)))
            else:
                return None

    except _NotNumerical:
        if len(branch._fLeaves) == 1:
            if len(branch._fBranches) > 0 and all(len(x._fLeaves) == 1 and x._fLeaves[0]._fLeafCount is branch._fLeaves[0] for x in branch._fBranches):
                return asdtype(">i4")

            if isinstance(branch._streamer, uproot.rootio.TStreamerObject):
                obj = uproot.rootio._safename(branch._streamer._fTypeName)
                if obj == "string":
                    return asgenobj(STLString(awkward), branch._context, 0)
                elif obj in branch._context.classes:
                    return _obj_or_genobj(branch._context.classes.get(obj), branch, isjagged, cntvers=cntvers, tobject=tobject, speedbump=speedbump)

            if isinstance(branch._streamer, uproot.rootio.TStreamerInfo):
                obj = uproot.rootio._safename(branch._streamer._fName)
                if obj == "string":
                    return asgenobj(STLString(awkward), branch._context, 0)
                elif obj in branch._context.classes:
                    return _obj_or_genobj(branch._context.classes.get(obj), branch, isjagged, cntvers=cntvers, tobject=tobject, speedbump=speedbump)

            if branch._fLeaves[0].__class__.__name__ == "TLeafC":
                return asstring(skipbytes=1)

            elif branch._fLeaves[0].__class__.__name__ == "TLeafElement":
                if isinstance(branch._streamer, uproot.rootio.TStreamerBasicType):
                    try:
                        fromdtype = _ftype2dtype(branch._streamer._fType, awkward)
                    except _NotNumerical:
                        pass
                    else:
                        if swapbytes:
                            todtype = fromdtype.newbyteorder("=")
                        else:
                            todtype = fromdtype
                        fromdims, remainder = divmod(branch._streamer._fSize, fromdtype.itemsize)
                        if remainder == 0:
                            todims = dims
                            if reduce(lambda x, y: x * y, todims, 1) != fromdims:
                                todims = (fromdims,)
                            return asdtype(awkward.numpy.dtype((fromdtype, (fromdims,))), awkward.numpy.dtype((todtype, todims)))

                if isinstance(branch._streamer, uproot.rootio.TStreamerBasicPointer):
                    if uproot.const.kOffsetP < branch._streamer._fType < uproot.const.kOffsetP + 20:
                        try:
                            fromdtype = _ftype2dtype(branch._streamer._fType - uproot.const.kOffsetP, awkward)
                        except _NotNumerical:
                            pass
                        else:
                            if swapbytes:
                                todtype = fromdtype.newbyteorder("=")
                            else:
                                todtype = fromdtype
                            if len(branch._fLeaves) == 1 and branch._fLeaves[0]._fLeafCount is not None:
                                return asjagged(asdtype(fromdtype, todtype), skipbytes=1)

                if isinstance(branch._streamer, uproot.rootio.TStreamerObjectAny):
                    if getattr(branch._streamer, "_fTypeName", None) in (b"TArrayC", b"TArrayS", b"TArrayI", b"TArrayL", b"TArrayL64", b"TArrayF", b"TArrayD"):
                        return asjagged(asdtype(getattr(uproot.rootio, branch._streamer._fTypeName.decode("ascii"))._dtype), skipbytes=4)

                if isinstance(branch._streamer, uproot.rootio.TStreamerString):
                    return asstring(skipbytes=1)

                if isinstance(branch._streamer, uproot.rootio.TStreamerSTLstring):
                    if branch._isTClonesArray:
                        return asgenobj(STLVector(STLString()), branch._context, 6)
                    else:
                        return asstring(skipbytes=7)

                if getattr(branch._streamer, "_fType", None) == uproot.const.kCharStar:
                    return asstring(skipbytes=4)

                if getattr(branch._streamer, "_fSTLtype", None) == uproot.const.kSTLvector:
                    try:
                        fromdtype = _ftype2dtype(branch._streamer._fCtype, awkward)
                        if swapbytes:
                            ascontent = asdtype(fromdtype, fromdtype.newbyteorder("="))
                        else:
                            ascontent = asdtype(fromdtype, fromdtype)
                        if branch._isTClonesArray:
                            return asgenobj(SimpleArray(STLVector(asdtype(fromdtype, fromdtype))), branch._context, 6)
                        else:
                            return asjagged(ascontent, skipbytes=10)

                    except _NotNumerical:
                        if branch._vecstreamer is not None:
                            try:
                                streamerClass = branch._vecstreamer.pyclass
                            except AttributeError:
                                obj = uproot.rootio._safename(branch._vecstreamer._fName)
                                if obj in branch._context.classes:
                                    streamerClass = branch._context.classes.get(obj)
                            if getattr(streamerClass, "_hasreadobjany", False):
                                return None

                            if streamerClass.__name__ == "string":
                                return asgenobj(STLVector(STLString(awkward)), branch._context, 6)

                            if len(branch._fBranches) != 0:
                                return None

                            try:
                                recarray = streamerClass._recarray_dtype(cntvers=cntvers, tobject=tobject)
                            except (AttributeError, ValueError):
                                return asgenobj(STLVector(streamerClass), branch._context, 6)
                            else:
                                if streamerClass._methods is None:
                                    return asjagged(astable(asdtype(recarray)), skipbytes=10)
                                else:
                                    return asjagged(asobj(astable(asdtype(recarray)), streamerClass._methods), skipbytes=10)

                if hasattr(branch._streamer, "_fTypeName"):
                    m = re.match(b"bitset<([1-9][0-9]*)>", branch._streamer._fTypeName)
                    if m is not None:
                        return asjagged(asstlbitset(int(m.group(1))), skipbytes=6)

                if getattr(branch._streamer, "_fTypeName", None) == b"vector<bool>" or getattr(branch._streamer, "_fTypeName", None) == b"vector<Bool_t>":
                    return asjagged(asdtype(awkward.numpy.bool_), skipbytes=10)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<char>" or getattr(branch._streamer, "_fTypeName", None) == b"vector<Char_t>":
                    return asjagged(asdtype("i1"), skipbytes=10)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<unsigned char>" or getattr(branch._streamer, "_fTypeName", None) == b"vector<UChar_t>" or getattr(branch._streamer, "_fTypeName", None) == b"vector<Byte_t>":
                    return asjagged(asdtype("u1"), skipbytes=10)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<short>" or getattr(branch._streamer, "_fTypeName", None) == b"vector<Short_t>":
                    return asjagged(asdtype("i2"), skipbytes=10)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<unsigned short>" or getattr(branch._streamer, "_fTypeName", None) == b"vector<UShort_t>":
                    return asjagged(asdtype("u2"), skipbytes=10)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<int>" or getattr(branch._streamer, "_fTypeName", None) == b"vector<Int_t>":
                    return asjagged(asdtype("i4"), skipbytes=10)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<unsigned int>" or getattr(branch._streamer, "_fTypeName", None) == b"vector<UInt_t>":
                    return asjagged(asdtype("u4"), skipbytes=10)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<long>" or getattr(branch._streamer, "_fTypeName", None) == b"vector<Long_t>":
                    return asjagged(asdtype("i8"), skipbytes=10)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<unsigned long>" or getattr(branch._streamer, "_fTypeName", None) == b"vector<ULong64_t>":
                    return asjagged(asdtype("u8"), skipbytes=10)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<float>" or getattr(branch._streamer, "_fTypeName", None) == b"vector<Float_t>":
                    return asjagged(asdtype("f4"), skipbytes=10)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<double>" or getattr(branch._streamer, "_fTypeName", None) == b"vector<Double_t>":
                    return asjagged(asdtype("f8"), skipbytes=10)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<string>":
                    return asgenobj(STLVector(STLString(awkward)), branch._context, 6)
                else:
                    m = interpret._vectorpointer.match(getattr(branch._streamer, "_fTypeName", b""))
                    if m is not None and m.group(1) in branch._context.streamerinfosmap:
                        streamer = branch._context.streamerinfosmap[m.group(1)]
                        return asgenobj(STLVector(Pointer(streamer.pyclass)), branch._context, skipbytes=6)

                if getattr(branch._streamer, "_fTypeName", None) == b"map<string,bool>" or getattr(branch._streamer, "_fTypeName", None) == b"map<string,Bool_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype(awkward.numpy.bool_)), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"map<string,char>" or getattr(branch._streamer, "_fTypeName", None) == b"map<string,Char_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("i1")), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"map<string,unsigned char>" or getattr(branch._streamer, "_fTypeName", None) == b"map<string,UChar_t>" or getattr(branch._streamer, "_fTypeName", None) == b"map<string,Byte_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("u1")), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"map<string,short>" or getattr(branch._streamer, "_fTypeName", None) == b"map<string,Short_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("i2")), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"map<string,unsigned short>" or getattr(branch._streamer, "_fTypeName", None) == b"map<string,UShort_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("u2")), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"map<string,int>" or getattr(branch._streamer, "_fTypeName", None) == b"map<string,Int_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("i4")), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"map<string,unsigned int>" or getattr(branch._streamer, "_fTypeName", None) == b"map<string,UInt_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("u4")), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"map<string,long>" or getattr(branch._streamer, "_fTypeName", None) == b"map<string,Long_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("i8")), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"map<string,unsigned long>" or getattr(branch._streamer, "_fTypeName", None) == b"map<string,ULong64_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("u8")), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"map<string,float>" or getattr(branch._streamer, "_fTypeName", None) == b"map<string,Float_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("f4")), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"map<string,double>" or getattr(branch._streamer, "_fTypeName", None) == b"map<string,Double_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("f8")), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"map<string,string>":
                    return asgenobj(STLMap(STLString(awkward), STLString(awkward)), branch._context, 6)

                if getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<bool> >" or getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<Bool_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(awkward.numpy.bool_))), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<char> >" or getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<Char_t> >":
                    return asgenobj(STLVector(STLVector(asdtype("i1"))), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<unsigned char> >" or getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<UChar_t> >" or getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<Byte_t> >":
                    return asgenobj(STLVector(STLVector(asdtype("u1"))), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<short> >" or getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<Short_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">i2"))), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<unsigned short> >" or getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<UShort_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">u2"))), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<int> >" or getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<Int_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">i4"))), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<unsigned int> >" or getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<UInt_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">u4"))), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<long> >" or getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<Long_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">i8"))), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<unsigned long> >" or getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<ULong64_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">u8"))), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<float> >" or getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<Float_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">f4"))), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<double> >" or getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<Double_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">f8"))), branch._context, 6)
                elif getattr(branch._streamer, "_fTypeName", None) == b"vector<vector<string> >":
                    return asgenobj(STLVector(STLVector(STLString(awkward))), branch._context, 6)

                m = re.match(b"bitset<([1-9][0-9]*)>", branch._fClassName)
                if m is not None:
                    return asstlbitset(int(m.group(1)))

                if branch._fClassName == b"string":
                    return asstring(skipbytes=1)

                if branch._fClassName == b"vector<bool>" or branch._fClassName == b"vector<Bool_t>":
                    return asjagged(asdtype(awkward.numpy.bool_), skipbytes=10)
                elif branch._fClassName == b"vector<char>" or branch._fClassName == b"vector<Char_t>":
                    return asjagged(asdtype("i1"), skipbytes=10)
                elif branch._fClassName == b"vector<unsigned char>" or branch._fClassName == b"vector<UChar_t>" or branch._fClassName == b"vector<Byte_t>":
                    return asjagged(asdtype("u1"), skipbytes=10)
                elif branch._fClassName == b"vector<short>" or branch._fClassName == b"vector<Short_t>":
                    return asjagged(asdtype("i2"), skipbytes=10)
                elif branch._fClassName == b"vector<unsigned short>" or branch._fClassName == b"vector<UShort_t>":
                    return asjagged(asdtype("u2"), skipbytes=10)
                elif branch._fClassName == b"vector<int>" or branch._fClassName == b"vector<Int_t>":
                    return asjagged(asdtype("i4"), skipbytes=10)
                elif branch._fClassName == b"vector<unsigned int>" or branch._fClassName == b"vector<UInt_t>":
                    return asjagged(asdtype("u4"), skipbytes=10)
                elif branch._fClassName == b"vector<long>" or branch._fClassName == b"vector<Long_t>":
                    return asjagged(asdtype("i8"), skipbytes=10)
                elif branch._fClassName == b"vector<unsigned long>" or branch._fClassName == b"vector<ULong64_t>":
                    return asjagged(asdtype("u8"), skipbytes=10)
                elif branch._fClassName == b"vector<float>" or branch._fClassName == b"vector<Float_t>":
                    return asjagged(asdtype("f4"), skipbytes=10)
                elif branch._fClassName == b"vector<double>" or branch._fClassName == b"vector<Double_t>":
                    return asjagged(asdtype("f8"), skipbytes=10)
                elif branch._fClassName == b"vector<string>":
                    return asgenobj(STLVector(STLString(awkward)), branch._context, 6)

                if branch._fClassName == b"vector<vector<bool> >" or branch._fClassName == b"vector<vector<Bool_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(awkward.numpy.bool_))), branch._context, 6)
                elif branch._fClassName == b"vector<vector<char> >" or branch._fClassName == b"vector<vector<Char_t> >":
                    return asgenobj(STLVector(STLVector(asdtype("i1"))), branch._context, 6)
                elif branch._fClassName == b"vector<vector<unsigned char> >" or branch._fClassName == b"vector<vector<UChar_t> >" or branch._fClassName == b"vector<vector<Byte_t> >":
                    return asgenobj(STLVector(STLVector(asdtype("u1"))), branch._context, 6)
                elif branch._fClassName == b"vector<vector<short> >" or branch._fClassName == b"vector<vector<Short_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">i2"))), branch._context, 6)
                elif branch._fClassName == b"vector<vector<unsigned short> >" or branch._fClassName == b"vector<vector<UShort_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">u2"))), branch._context, 6)
                elif branch._fClassName == b"vector<vector<int> >" or branch._fClassName == b"vector<vector<Int_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">i4"))), branch._context, 6)
                elif branch._fClassName == b"vector<vector<unsigned int> >" or branch._fClassName == b"vector<vector<UInt_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">u4"))), branch._context, 6)
                elif branch._fClassName == b"vector<vector<long> >" or branch._fClassName == b"vector<vector<Long_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">i8"))), branch._context, 6)
                elif branch._fClassName == b"vector<vector<unsigned long> >" or branch._fClassName == b"vector<vector<ULong64_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">u8"))), branch._context, 6)
                elif branch._fClassName == b"vector<vector<float> >" or branch._fClassName == b"vector<vector<Float_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">f4"))), branch._context, 6)
                elif branch._fClassName == b"vector<vector<double> >" or branch._fClassName == b"vector<vector<Double_t> >":
                    return asgenobj(STLVector(STLVector(asdtype(">f8"))), branch._context, 6)
                elif branch._fClassName == b"vector<vector<string> >":
                    return asgenobj(STLVector(STLVector(STLString(awkward))), branch._context, 6)

                if branch._fClassName == b"map<string,bool>" or branch._fClassName == b"map<string,Bool_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype(awkward.numpy.bool_)), branch._context, 6)
                elif branch._fClassName == b"map<string,char>" or branch._fClassName == b"map<string,Char_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("i1")), branch._context, 6)
                elif branch._fClassName == b"map<string,unsigned char>" or branch._fClassName == b"map<string,UChar_t>" or branch._fClassName == b"map<string,Byte_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("u1")), branch._context, 6)
                elif branch._fClassName == b"map<string,short>" or branch._fClassName == b"map<string,Short_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("i2")), branch._context, 6)
                elif branch._fClassName == b"map<string,unsigned short>" or branch._fClassName == b"map<string,UShort_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("u2")), branch._context, 6)
                elif branch._fClassName == b"map<string,int>" or branch._fClassName == b"map<string,Int_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("i4")), branch._context, 6)
                elif branch._fClassName == b"map<string,unsigned int>" or branch._fClassName == b"map<string,UInt_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("u4")), branch._context, 6)
                elif branch._fClassName == b"map<string,long>" or branch._fClassName == b"map<string,Long_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("i8")), branch._context, 6)
                elif branch._fClassName == b"map<string,unsigned long>" or branch._fClassName == b"map<string,ULong64_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("u8")), branch._context, 6)
                elif branch._fClassName == b"map<string,float>" or branch._fClassName == b"map<string,Float_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("f4")), branch._context, 6)
                elif branch._fClassName == b"map<string,double>" or branch._fClassName == b"map<string,Double_t>":
                    return asgenobj(STLMap(STLString(awkward), asdtype("f8")), branch._context, 6)
                elif branch._fClassName == b"map<string,string>":
                    return asgenobj(STLMap(STLString(awkward), STLString(awkward)), branch._context, 6)

                if branch.name.endswith(b".first") and branch._fClassName.startswith(b"pair<string,"):
                    return asgenobj(SimpleArray(STLString(awkward)), branch._context, 6)

                if branch.name.endswith(b".second"):
                    m = interpret._pairsecond.match(branch._fClassName)
                    if m is not None:
                        t, = m.groups()
                        if t == b"vector<bool>" or t == b"vector<Bool_t>":
                            return asgenobj(SimpleArray(STLVector(asdtype(awkward.numpy.bool_))), branch._context, 6)
                        elif t == b"vector<char>" or t == b"vector<Char_t>":
                            return asgenobj(SimpleArray(STLVector(asdtype("i1"))), branch._context, 6)
                        elif t == b"vector<unsigned char>" or t == b"vector<UChar_t>" or t == b"vector<Byte_t>":
                            return asgenobj(SimpleArray(STLVector(asdtype("u1"))), branch._context, 6)
                        elif t == b"vector<short>" or t == b"vector<Short_t>":
                            return asgenobj(SimpleArray(STLVector(asdtype("i2"))), branch._context, 6)
                        elif t == b"vector<unsigned short>" or t == b"vector<UShort_t>":
                            return asgenobj(SimpleArray(STLVector(asdtype("u2"))), branch._context, 6)
                        elif t == b"vector<int>" or t == b"vector<Int_t>":
                            return asgenobj(SimpleArray(STLVector(asdtype("i4"))), branch._context, 6)
                        elif t == b"vector<unsigned int>" or t == b"vector<UInt_t>":
                            return asgenobj(SimpleArray(STLVector(asdtype("u4"))), branch._context, 6)
                        elif t == b"vector<long>" or t == b"vector<Long_t>":
                            return asgenobj(SimpleArray(STLVector(asdtype("i8"))), branch._context, 6)
                        elif t == b"vector<unsigned long>" or t == b"vector<ULong64_t>":
                            return asgenobj(SimpleArray(STLVector(asdtype("u8"))), branch._context, 6)
                        elif t == b"vector<float>" or t == b"vector<Float_t>":
                            return asgenobj(SimpleArray(STLVector(asdtype("f4"))), branch._context, 6)
                        elif t == b"vector<double>" or t == b"vector<Double_t>":
                            return asgenobj(SimpleArray(STLVector(asdtype("f8"))), branch._context, 6)
                        elif t == b"vector<string>":
                            return asgenobj(SimpleArray(STLVector(STLString(awkward))), branch._context, 6)

        return None

interpret._titlehasdims = re.compile(br"^([^\[\]]+)(\[[^\[\]]+\])+")
interpret._itemdimpattern = re.compile(br"\[([1-9][0-9]*)\]")
interpret._itemanypattern = re.compile(br"\[(.*)\]")
interpret._vectorpointer = re.compile(br"vector\<([^<>]*)\*\>")
interpret._pairsecond = re.compile(br"pair\<[^<>]*,(.*) \>")

streamer_aliases = [
    (re.compile(b'(ROOT::Math::(?:PositionVector3D|DisplacementVector3D)<ROOT::Math::Cartesian3D<(?:[^>,]+)>)\\s+(>)'),
        b'\\1,ROOT::Math::DefaultCoordinateSystemTag\\2'),
]



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot/blob/master/LICENSE

"""Top-level functions for Pandas."""
from __future__ import absolute_import

import uproot.tree
from uproot.source.memmap import MemmapSource
from uproot.source.xrootd import XRootDSource
from uproot.source.http import HTTPSource

def iterate(path, treepath, branches=None, entrysteps=None, namedecode="utf-8", reportpath=False, reportfile=False, flatten=True, flatname=None, awkwardlib=None, cache=None, basketcache=None, keycache=None, executor=None, blocking=True, localsource=MemmapSource.defaults, xrootdsource=XRootDSource.defaults, httpsource=HTTPSource.defaults, **options):
    import pandas
    return uproot.tree.iterate(path, treepath, branches=branches, entrysteps=entrysteps, outputtype=pandas.DataFrame, namedecode=namedecode, reportpath=reportpath, reportfile=reportfile, reportentries=False, flatten=flatten, flatname=flatname, awkwardlib=awkwardlib, cache=cache, basketcache=basketcache, keycache=keycache, executor=executor, blocking=blocking, localsource=localsource, xrootdsource=xrootdsource, httpsource=httpsource, **options)



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot/blob/master/LICENSE

from __future__ import absolute_import

import datetime

def datime(when=None):
    if when is None:
        when = datetime.datetime.now()
    return (when.year - 1995) << 26 | when.month << 22 | when.day << 17 | when.hour << 12 | when.minute << 6 | when.second



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot/blob/master/LICENSE

"""uproot -- ROOT I/O in pure Python and Numpy.

Basic cheat-sheet
-----------------

Open ROOT files with uproot.open (for reading) or uproot.create (for read-write).

    file = uproot.open("/path/to/my/file.root")
    file = uproot.open("root://path/to/my/file.root")
    file = uproot.open("http://path/to/my/file.root")
    writeable = uproot.create("/new/local/file.root")

These file objects act like dicts; get objects like TTrees from them with square
brackets.

    tree = file["path/to/events"]
    tree = file["path/to/events;2"]            # optional cycle number
    # write to files by assignment (histograms only)
    writeable["name"] = numpy.histogram(...)

TTree objects also act like dicts; get branches with square brackets or list them
with keys().

    tree.keys()
    tree.allkeys()    # recursive branches-of-branches
    tree.show()       # display view
    tree["mybranch"]  # searches recursively

Get data as arrays with an array(...) or arrays(...) call.

    tree["mybranch"].array()
    tree.array("mybranch")
    tree.arrays(["branch1", "branch2", "branch3"])
    tree.arrays(["Muon_*"])

Variable numbers of objects per entry (particles per event) are handled by
awkward-array:

    https://github.com/scikit-hep/awkward-array

The arrays(...) call returns a dict from branch name (bytes) to data
(Numpy array) by default.
Change this by passing an outputtype class (e.g. dict, tuple, pandas.DataFrame).

    x, y, z = tree.arrays(["x", "y", "z"], outputtype=tuple)

For more idiomatic Pandas defaults, use tree.pandas.df().

    df = tree.pandas.df()

If the desired branches do not fit into memory, iterate over chunks of entries
with iterate(). The interface is the same as above: you get the same
dict/tuple/DataFrame with fewer entries.

    for x, y, z in tree.iterate(["x", "y", "z"], outputtype=tuple):
        do_something(x, y, z)

To iterate over many files (like TChain), do uproot.iterate(...).

    for arrays in uproot.iterate("files*.root", "path/to/events", ["Muon_*"]):
        do_something(arrays)

Intermediate cheat-sheet
------------------------

Each call to array/arrays/iterate reads the file again. For faster access after
the first time, pass a dict-like object to the cache parameter and uproot will
try the cache first.

    cache = {}
    arrays = tree.arrays(["Muon_*"], cache=cache)    # slow
    arrays = tree.arrays(["Muon_*"], cache=cache)    # fast

You control the cache object. If you're running out of memory, remove it or
remove items from it. Or use one of the dict-like caches from cachetools (already
installed) or another library.

For parallel processing, pass a Python 3 executor.

    import concurrent.futures
    executor = concurrent.futures.ThreadPoolExecutor(32)
    arrays = tree.arrays(["Muon_*"], executor=executor)

To get the number of entries per file in a a collection of files, use
uproot.numentries().

    uproot.numentries("tests/samples/sample*.root", "sample", total=False)

For arrays that read on demand, use uproot.lazyarray and uproot.lazyarrays.
For processing with Dask, use uproot.daskarray, uproot.daskarrays, or
uproot.daskframe.

Advanced cheat-sheet
--------------------

The standard bytes-to-arrays decoding is attached to each branch as

    tree["mybranch"].interpretation

This can be overridden by passing a new interpretation to array/arrays/iterate.
Most reinterpretations will produce wrong values (it's a reinterpret_cast<...>).

Some, however, are useful:

    mybranch = tree["mybranch"]
    fill_me_instead = numpy.empty(big_enough)
    mybranch.array(mybranch.interpretation.toarray(fill_me_instead))
    fill_me_instead                   # filled in place

    mybranch.array(uproot.asdebug)    # view raw bytes of each entry

By default, local files are read as memory-mapped arrays. Change this by setting

    from uproot import FileSource
    open("...", localsource=lambda path: FileSource(path, **FileSource.defaults))

The same procedure sets options for uproot.XRootDSource and uproot.HTTPSource.
"""

from __future__ import absolute_import

# high-level entry points
from uproot.rootio import open, xrootd, http
from uproot.tree import iterate, numentries, lazyarray, lazyarrays, daskarray, daskframe
from uproot.write.TFile import TFileCreate as create
from uproot.write.TFile import TFileRecreate as recreate
from uproot.write.TFile import TFileUpdate as update
from uproot.write.compress import ZLIB, LZMA, LZ4
from uproot.write.objects.TTree import newtree, newbranch

from uproot.source.memmap import MemmapSource
from uproot.source.file import FileSource
from uproot.source.xrootd import XRootDSource
from uproot.source.http import HTTPSource

from uproot.cache import ArrayCache, ThreadSafeArrayCache

from uproot.interp.auto import interpret
from uproot.interp.numerical import asdtype
from uproot.interp.numerical import asarray
from uproot.interp.numerical import asdouble32
from uproot.interp.numerical import asstlbitset
from uproot.interp.jagged import asjagged
from uproot.interp.objects import astable
from uproot.interp.objects import asobj
from uproot.interp.objects import asgenobj
from uproot.interp.objects import asstring
from uproot.interp.objects import SimpleArray
from uproot.interp.objects import STLVector
from uproot.interp.objects import STLMap
from uproot.interp.objects import STLString
from uproot.interp.objects import Pointer
asdebug = asjagged(asdtype("u1"))

from uproot import pandas

# put help strings on everything (they're long, too disruptive to intersperse
# in the code, and are built programmatically to avoid duplication; Python's
# inline docstring method doesn't accept non-literals)
import uproot._help

# convenient access to the version number
from uproot.version import __version__

# don't expose uproot.uproot; it's ugly
del uproot

__all__ = ["open", "xrootd", "http", "iterate", "numentries", "lazyarray", "lazyarrays", "daskarray", "daskframe", "create", "recreate", "update", "ZLIB", "LZMA", "LZ4", "ZSTD", "newtree", "newbranch", "MemmapSource", "FileSource", "XRootDSource", "HTTPSource", "ArrayCache", "ThreadSafeArrayCache", "interpret", "asdtype", "asarray", "asdouble32", "asstlbitset", "asjagged", "astable", "asobj", "asgenobj", "asstring", "asdebug", "SimpleArray", "STLVector", "STLMap", "STLString", "Pointer", "pandas", "__version__"]



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot/blob/master/LICENSE

import uproot

class Test(object):
    def test_flat_array(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["i8"]
        expectation = [-15, -14, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]

        cache = {}
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert len(cache) == 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, cache=cache).tolist() == expectation[entrystart:entrystop]
            assert len(cache) > 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, cache=cache).tolist() == expectation[entrystart:entrystop]
            cache = {}

        basketcache = {}
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert len(basketcache) == 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, basketcache=basketcache).tolist() == expectation[entrystart:entrystop]
            assert len(basketcache) > 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, basketcache=basketcache).tolist() == expectation[entrystart:entrystop]
            basketcache = {}

        keycache = {}
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert len(keycache) == 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, keycache=keycache).tolist() == expectation[entrystart:entrystop]
            assert len(keycache) > 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, keycache=keycache).tolist() == expectation[entrystart:entrystop]
            keycache = {}

    def test_regular_array(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["ai8"]
        expectation = [[-14, -13, -12], [-13, -12, -11], [-12, -11, -10], [-11, -10, -9], [-10, -9, -8], [-9, -8, -7], [-8, -7, -6], [-7, -6, -5], [-6, -5, -4], [-5, -4, -3], [-4, -3, -2], [-3, -2, -1], [-2, -1, 0], [-1, 0, 1], [0, 1, 2], [1, 2, 3], [2, 3, 4], [3, 4, 5], [4, 5, 6], [5, 6, 7], [6, 7, 8], [7, 8, 9], [8, 9, 10], [9, 10, 11], [10, 11, 12], [11, 12, 13], [12, 13, 14], [13, 14, 15], [14, 15, 16], [15, 16, 17]]

        cache = {}
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert len(cache) == 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, cache=cache).tolist() == expectation[entrystart:entrystop]
            assert len(cache) > 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, cache=cache).tolist() == expectation[entrystart:entrystop]
            cache = {}

        basketcache = {}
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert len(basketcache) == 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, basketcache=basketcache).tolist() == expectation[entrystart:entrystop]
            assert len(basketcache) > 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, basketcache=basketcache).tolist() == expectation[entrystart:entrystop]
            basketcache = {}

        keycache = {}
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert len(keycache) == 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, keycache=keycache).tolist() == expectation[entrystart:entrystop]
            assert len(keycache) > 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, keycache=keycache).tolist() == expectation[entrystart:entrystop]
            keycache = {}

    def test_irregular_array(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["Ai8"]
        expectation = [[], [-15], [-15, -13], [-15, -13, -11], [-15, -13, -11, -9], [], [-10], [-10, -8], [-10, -8, -6], [-10, -8, -6, -4], [], [-5], [-5, -3], [-5, -3, -1], [-5, -3, -1, 1], [], [0], [0, 2], [0, 2, 4], [0, 2, 4, 6], [], [5], [5, 7], [5, 7, 9], [5, 7, 9, 11], [], [10], [10, 12], [10, 12, 14], [10, 12, 14, 16]]
        assert [len(x) for x in expectation] == [0, 1, 2, 3, 4] * 6

        cache = {}
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert len(cache) == 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, cache=cache).tolist() == expectation[entrystart:entrystop]
            assert len(cache) > 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, cache=cache).tolist() == expectation[entrystart:entrystop]
            cache = {}

        basketcache = {}
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert len(basketcache) == 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, basketcache=basketcache).tolist() == expectation[entrystart:entrystop]
            assert len(basketcache) > 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, basketcache=basketcache).tolist() == expectation[entrystart:entrystop]
            basketcache = {}

        keycache = {}
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert len(keycache) == 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, keycache=keycache).tolist() == expectation[entrystart:entrystop]
            assert len(keycache) > 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, keycache=keycache).tolist() == expectation[entrystart:entrystop]
            keycache = {}

    def test_strings_array(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["str"]
        expectation = [b"hey-0", b"hey-1", b"hey-2", b"hey-3", b"hey-4", b"hey-5", b"hey-6", b"hey-7", b"hey-8", b"hey-9", b"hey-10", b"hey-11", b"hey-12", b"hey-13", b"hey-14", b"hey-15", b"hey-16", b"hey-17", b"hey-18", b"hey-19", b"hey-20", b"hey-21", b"hey-22", b"hey-23", b"hey-24", b"hey-25", b"hey-26", b"hey-27", b"hey-28", b"hey-29"]

        cache = {}
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert len(cache) == 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, cache=cache).tolist() == expectation[entrystart:entrystop]
            assert len(cache) > 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, cache=cache).tolist() == expectation[entrystart:entrystop]
            cache = {}

        basketcache = {}
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert len(basketcache) == 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, basketcache=basketcache).tolist() == expectation[entrystart:entrystop]
            assert len(basketcache) > 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, basketcache=basketcache).tolist() == expectation[entrystart:entrystop]
            basketcache = {}

        keycache = {}
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert len(keycache) == 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, keycache=keycache).tolist() == expectation[entrystart:entrystop]
            assert len(keycache) > 0
            assert branch.array(entrystart=entrystart, entrystop=entrystop, keycache=keycache).tolist() == expectation[entrystart:entrystop]
            keycache = {}



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot/blob/master/LICENSE

from __future__ import absolute_import

import struct

import uproot.write.sink.cursor
import uproot.write.util

class BasketKey(object):
    def __init__(self, fName, fTitle, fNevBuf, fNevBufSize, fObjlen=0, fSeekKey=100, fSeekPdir=0, fBufferSize=0):
        self.fClassName = b"TBasket"
        self.fName = fName
        self.fTitle = fTitle

        self.fObjlen = fObjlen
        self.fSeekKey = fSeekKey
        self.fSeekPdir = fSeekPdir
        self.fCycle = 0
        self.fDatime = uproot.write.util.datime()
        self.fNbytes = self.fObjlen + self.fKeylen
        self.fBufferSize = fBufferSize
        self.fNevBuf = fNevBuf
        self.fNevBufSize = fNevBufSize

        self.old_fLast = 0

    @property
    def fKeylen(self):
        return self._format1.size + uproot.write.sink.cursor.Cursor.length_strings([self.fClassName, self.fName, self.fTitle]) + self._format_basketkey.size + 1

    @property
    def fLast(self):
        return self.fKeylen + self.fObjlen

    def update(self):
        self.cursor.update_fields(self.sink, self._format1, self.fNbytes, self._version, self.fObjlen, self.fDatime, self.fKeylen, self.fCycle, self.fSeekKey, self.fSeekPdir)

    def write(self, cursor, sink, isjagged=False):
        self.cursor = uproot.write.sink.cursor.Cursor(cursor.index)
        self.sink = sink

        self.update()

        cursor.skip(self._format1.size)
        cursor.write_string(sink, self.fClassName)
        cursor.write_string(sink, self.fName)
        cursor.write_string(sink, self.fTitle)

        basketversion = 3
        if isjagged:
            if self.old_fLast == 0:
                raise Exception("isjagged flag should be False")
            cursor.write_fields(sink, self._format_basketkey, basketversion, self.fBufferSize, self.fNevBufSize, self.fNevBuf, self.old_fLast)
        else:
            cursor.write_fields(sink, self._format_basketkey, basketversion, self.fBufferSize, self.fNevBufSize, self.fNevBuf, self.fLast)
            self.old_fLast = self.fLast
        cursor.write_data(sink, b"\x00")

    _version = 1004
    _format1 = struct.Struct(">ihiIhhqq")
    _format_basketkey = struct.Struct(">Hiiii")

class TKey(object):
    def __init__(self, fClassName, fName, fTitle=b"", fObjlen=0, fSeekKey=100, fSeekPdir=0, fCycle=1):
        self.fClassName = fClassName
        self.fName = fName
        self.fTitle = fTitle

        self.fObjlen = fObjlen
        self.fSeekKey = fSeekKey
        self.fSeekPdir = fSeekPdir
        self.fCycle = fCycle
        self.fDatime = uproot.write.util.datime()
        self.fNbytes = self.fObjlen + self.fKeylen

    @property
    def fKeylen(self):
        return self._format1.size + uproot.write.sink.cursor.Cursor.length_strings([self.fClassName, self.fName, self.fTitle])

    def update(self):
        self.cursor.update_fields(self.sink, self._format1, self.fNbytes, self._version, self.fObjlen, self.fDatime, self.fKeylen, self.fCycle, self.fSeekKey, self.fSeekPdir)

    def write(self, cursor, sink, isjagged=False):
        if isjagged:
            raise Exception("isjagged flag should be False")
        self.cursor = uproot.write.sink.cursor.Cursor(cursor.index)
        self.sink = sink

        self.update()

        cursor.skip(self._format1.size)
        cursor.write_string(sink, self.fClassName)
        cursor.write_string(sink, self.fName)
        cursor.write_string(sink, self.fTitle)

    _version = 1004
    _format1 = struct.Struct(">ihiIhhqq")

class TKey32(TKey):
    _version = 4
    _format1 = struct.Struct(">ihiIhhii")



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot/blob/master/LICENSE

from __future__ import absolute_import

import math

import numpy

import uproot.cache
import uproot.source.source


class ChunkedSource(uproot.source.source.Source):
    # makes __doc__ attribute mutable before Python 3.3
    __metaclass__ = type.__new__(type, "type", (uproot.source.source.Source.__metaclass__,), {})

    def __init__(self, path, chunkbytes, limitbytes, parallel):
        from uproot.rootio import _memsize
        m = _memsize(chunkbytes)
        if m is not None:
            chunkbytes = int(math.ceil(m))
        m = _memsize(limitbytes)
        if m is not None:
            limitbytes = int(math.ceil(m))
        self.path = path
        self._chunkbytes = chunkbytes
        self._limitbytes = limitbytes
        if limitbytes is None:
            self.cache = {}
        else:
            self.cache = uproot.cache.ThreadSafeArrayCache(limitbytes)
        self._source = None
        self._setup_futures(parallel)

    def parent(self):
        return self

    def threadlocal(self):
        return self

    def __del__(self):
        self.dismiss()

    def _read(self, chunkindex):
        raise NotImplementedError

    def close(self):
        super(ChunkedSource, self).close()
        self.cache.clear()

    def dismiss(self):
        if self._futures is not None:
            for future in self._futures.values():
                future.cancel()
            self._futures = {}

    def _setup_futures(self, parallel):
        if parallel is not None and parallel > 1:
            try:
                import concurrent.futures
            except ImportError:
                raise ImportError("Install futures package (for parallel > 1) with:\n    pip install futures\nor\n    conda install -c conda-forge futures")
            self._executor = concurrent.futures.ThreadPoolExecutor(parallel)
            self._futures = {}
        else:
            self._executor = None
            self._futures = None

    def _preload(self, chunkindex):
        try:
            chunk = self.cache[chunkindex]
        except KeyError:
            return self._read(chunkindex)
        else:
            return chunk

    def preload(self, starts):
        self._open()
        limitnum = self._limitbytes // self._chunkbytes
        if self._executor is not None:
            for start in starts:
                if len(self._futures) > limitnum:
                    break
                chunkindex = start // self._chunkbytes
                if chunkindex not in self._futures:
                    self._futures[chunkindex] = self._executor.submit(self._preload, chunkindex)

    def data(self, start, stop, dtype=None):
        if dtype is None:
            thedtype = numpy.dtype(numpy.uint8)
        else:
            thedtype = dtype

        # assert start >= 0
        # assert stop >= 0
        # assert stop >= start

        chunkstart = start // self._chunkbytes
        if stop % self._chunkbytes == 0:
            chunkstop = stop // self._chunkbytes
        else:
            chunkstop = stop // self._chunkbytes + 1

        out = numpy.empty((stop - start) // thedtype.itemsize, dtype=thedtype)

        for chunkindex in range(chunkstart, chunkstop):
            chunk = None
            if self._futures is not None:
                future = self._futures.pop(chunkindex, None)
                if future is not None:
                    chunk = future.result()

            if chunk is None:
                try:
                    chunk = self.cache[chunkindex]
                except KeyError:
                    self._open()
                    chunk = self._read(chunkindex)

            cstart = 0
            cstop = self._chunkbytes
            gstart = chunkindex * self._chunkbytes
            gstop = (chunkindex + 1) * self._chunkbytes

            if len(chunk) > self._chunkbytes:
                if not numpy.array_equal(chunk[:4], list(b"root")):
                    raise NotImplementedError("Expected {0} or fewer bytes but received {1} and data does not appear to be an entire ROOT file.".format(self._chunkbytes, len(chunk)))
                self.cache = {}
                for i in range(0, len(chunk), self._chunkbytes):
                    self.cache[i // self._chunkbytes] = chunk[i:i+self._chunkbytes]
                chunk = self.cache[chunkindex]
                # Dismiss any pending futures as everything has already been loaded
                self.dismiss()
            else:
                self.cache[chunkindex] = chunk

            if gstart < start:
                cstart += start - gstart
                gstart += start - gstart
            if gstop > stop:
                cstop -= gstop - stop
                gstop -= gstop - stop

            if cstop - cstart > len(chunk):
                raise IndexError("indexes {0}:{1} are beyond the end of data source {2}".format(gstart + len(chunk), stop, repr(self.path)))

            if dtype is None:
                out[gstart - start : gstop - start] = chunk[cstart:cstop]
            else:
                out.view(numpy.uint8)[gstart - start : gstop - start] = chunk[cstart:cstop]

        return out



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot/blob/master/LICENSE

from __future__ import absolute_import

import struct
from copy import copy

import numpy

import uproot.const
import uproot.write.compress
import uproot.write.sink.cursor
from uproot.rootio import _bytesid

class TH(object):
    def __init__(self, histogram):
        self._fTitle = _bytesid(histogram._fTitle)
        self._fClassName = _bytesid(histogram._classname)

        self._fXaxis = self._emptyaxis("xaxis", 1.0)
        self._fXaxis.update(histogram._fXaxis.__dict__)
        self._fixaxis(self._fXaxis)

        self._fYaxis = self._emptyaxis("yaxis", 0.0)
        if hasattr(histogram, "_fYaxis"):
            self._fYaxis.update(histogram._fYaxis.__dict__)
        self._fixaxis(self._fYaxis)

        self._fZaxis = self._emptyaxis("zaxis", 1.0)
        if hasattr(histogram, "_fZaxis"):
            self._fZaxis.update(histogram._fZaxis.__dict__)
        self._fixaxis(self._fZaxis)

        self._values = histogram.allvalues

        self._fields = self._emptyfields()
        for n in list(self._fields):
            if hasattr(histogram, n):
                self._fields[n] = getattr(histogram, n)

        if self._fClassName == b"TH1C":
            self._valuesarray = numpy.array(self._values, dtype=">i1")
        elif self._fClassName == b"TH1S":
            self._valuesarray = numpy.array(self._values, dtype=">i2")
        elif self._fClassName == b"TH1I":
            self._valuesarray = numpy.array(self._values, dtype=">i4")
        elif self._fClassName == b"TH1F":
            self._valuesarray = numpy.array(self._values, dtype=">f4")
        elif self._fClassName == b"TH1D":
            self._valuesarray = numpy.array(self._values, dtype=">f8")
        elif self._fClassName == b"TProfile":
            self._valuesarray = numpy.array(self._values, dtype=">f8")
        elif self._fClassName == b"TH2C":
            self._valuesarray = numpy.array(self._values, dtype=">i1").transpose()
        elif self._fClassName == b"TH2S":
            self._valuesarray = numpy.array(self._values, dtype=">i2").transpose()
        elif self._fClassName == b"TH2I":
            self._valuesarray = numpy.array(self._values, dtype=">i4").transpose()
        elif self._fClassName == b"TH2F":
            self._valuesarray = numpy.array(self._values, dtype=">f4").transpose()
        elif self._fClassName == b"TH2D":
            self._valuesarray = numpy.array(self._values, dtype=">f8").transpose()
        elif self._fClassName == b"TProfile2D":
            self._valuesarray = numpy.array(self._values, dtype=">f8").transpose()
        elif self._fClassName == b"TH3C":
            self._valuesarray = numpy.array(self._values, dtype=">i1").transpose()
        elif self._fClassName == b"TH3S":
            self._valuesarray = numpy.array(self._values, dtype=">i2").transpose()
        elif self._fClassName == b"TH3I":
            self._valuesarray = numpy.array(self._values, dtype=">i4").transpose()
        elif self._fClassName == b"TH3F":
            self._valuesarray = numpy.array(self._values, dtype=">f4").transpose()
        elif self._fClassName == b"TH3D":
            self._valuesarray = numpy.array(self._values, dtype=">f8").transpose()
        elif self._fClassName == b"TProfile3D":
            self._valuesarray = numpy.array(self._values, dtype=">f8").transpose()
        else:
            raise ValueError("unrecognized histogram class name {0}".format(self._fClassName))

        self._fields["_fNcells"] = self._valuesarray.size
        self._fields["_fContour"] = numpy.array(self._fields["_fContour"], dtype=">f8", copy=False)
        self._fields["_fSumw2"] = numpy.array(self._fields["_fSumw2"], dtype=">f8", copy=False)
        self._fields["_fBinEntries"] = numpy.array(self._fields["_fBinEntries"], dtype=">f8", copy=False)
        self._fields["_fBinSumw2"] = numpy.array(self._fields["_fBinSumw2"], dtype=">f8", copy=False)

    @staticmethod
    def _fixaxis(axis):
        axis["_fName"] = _bytesid(axis["_fName"])
        axis["_fTitle"] = _bytesid(axis["_fTitle"])
        axis["_fXbins"] = numpy.array(axis["_fXbins"], dtype=">f8", copy=False)
        if axis["_fLabels"] is None:
            axis["_fLabels"] = []
        if axis["_fModLabs"] is None:
            axis["_fModLabs"] = []

    @staticmethod
    def _emptyfields():
        return {"_fLineColor": 602,
                "_fLineStyle": 1,
                "_fLineWidth": 1,
                "_fFillColor": 0,
                "_fFillStyle": 1001,
                "_fMarkerColor": 1,
                "_fMarkerStyle": 1,
                "_fMarkerSize": 1.0,
                "_fNcells": 12,
                "_fBarOffset": 0,
                "_fBarWidth": 1000,
                "_fEntries": 0.0,
                "_fTsumw": 0.0,
                "_fTsumw2": 0.0,
                "_fTsumwx": 0.0,
                "_fTsumwx2": 0.0,
                "_fMaximum": -1111.0,
                "_fMinimum": -1111.0,
                "_fNormFactor": 0.0,
                "_fContour": [],
                "_fSumw2": [],
                "_fOption": b"",
                "_fFunctions": [],
                "_fBufferSize": 0,
                "_fBuffer": [],
                "_fBinStatErrOpt": 0,
                "_fStatOverflows": 2,
                "_fTsumwy": 0.0,
                "_fTsumwy2": 0.0,
                "_fTsumwxy": 0.0,
                "_fTsumwz": 0.0,
                "_fTsumwz2": 0.0,
                "_fTsumwxz": 0.0,
                "_fTsumwyz": 0.0,
                "_fScalefactor": 0.0,
                "_fBinEntries": [],
                "_fYmin": 0.0,
                "_fYmax": 0.0,
                "_fBinSumw2": [],
                "_fErrorMode": 0,
                "_fZmin": 0.0,
                "_fZmax": 0.0,
                "_fTmin": 0.0,
                "_fTmax": 0.0,
                "_fTsumwt": 0.0,
                "_fTsumwt2": 0.0}

    @staticmethod
    def _emptyaxis(name, titleoffset):
        return {"_fName": name,
                "_fTitle": b"",
                "_fNdivisions": 510,
                "_fAxisColor": 1,
                "_fLabelColor": 1,
                "_fLabelFont": 42,
                "_fLabelOffset": 0.004999999888241291,
                "_fLabelSize": 0.03500000014901161,
                "_fTickLength": 0.029999999329447746,
                "_fTitleOffset": titleoffset,
                "_fTitleSize": 0.03500000014901161,
                "_fTitleColor": 1,
                "_fTitleFont": 42,
                "_fNbins": 1,
                "_fXmin": 0.0,
                "_fXmax": 1.0,
                "_fXbins": [],
                "_fFirst": 0,
                "_fLast": 0,
                "_fBits2": 0,
                "_fTimeDisplay": False,
                "_fTimeFormat": b"",
                "_fLabels": None,
                "_fModLabs": None}

    _format_cntvers = struct.Struct(">IH")

    _format_tobject1 = struct.Struct(">HII")
    def _put_tobject(self, cursor):
        return cursor.put_fields(self._format_tobject1, 1, 0, numpy.uint32(0x03000000))

    def _put_tnamed(self, cursor, name, title):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        vers = 1
        buff = (self._put_tobject(cursor) +
                cursor.put_string(name) + cursor.put_string(title))
        length = len(buff) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers) + buff

    _format_tarray = struct.Struct(">i")
    def _put_tarray(self, cursor, values):
        return cursor.put_fields(self._format_tarray, values.size) + cursor.put_array(values)

    _format_tobjstring = struct.Struct(">IHHII")
    def _put_tobjstring(self, cursor, value, bit=0):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_tobjstring.size)
        vers = 1
        buff = cursor.put_string(value)
        length = len(buff) + self._format_tobjstring.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_tobjstring, cnt, vers, 1, bit, numpy.uint32(0x03000000)) + buff

    _format_tlist1 = struct.Struct(">i")
    _format_tlist2 = struct.Struct(">B")
    def _put_tlist(self, cursor, values):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        vers = 5
        buff = b""
        givenbytes = (self._put_tobject(cursor) +
                      cursor.put_string(b"") +
                      cursor.put_fields(self._format_tlist1, len(values)))
        for value in values:
            buff += self.util.put_objany(cursor, (value, "TObjString"), self.keycursor)
            buff += cursor.put_fields(self._format_tlist2, 0)
            buff += b"" # cursor.bytes(source, n)
        givenbytes += buff
        length = len(givenbytes) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers) + givenbytes

    _format_tattline = struct.Struct(">hhh")
    def _put_tattline(self, cursor):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        vers = 2
        buff = (cursor.put_fields(self._format_tattline,
                                  self._fields["_fLineColor"],
                                  self._fields["_fLineStyle"],
                                  self._fields["_fLineWidth"]))
        length = len(buff) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers) + buff

    _format_tattfill = struct.Struct(">hh")
    def _put_tattfill(self, cursor):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        vers = 2
        buff = (cursor.put_fields(self._format_tattfill,
                                  self._fields["_fFillColor"],
                                  self._fields["_fFillStyle"]))
        length = len(buff) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers) + buff

    _format_tattmarker = struct.Struct(">hhf")
    def _put_tattmarker(self, cursor):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        vers = 2
        buff = (cursor.put_fields(self._format_tattmarker,
                                  self._fields["_fMarkerColor"],
                                  self._fields["_fMarkerStyle"],
                                  self._fields["_fMarkerSize"]))
        length = len(buff) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers) + buff

    _format_tattaxis = struct.Struct(">ihhhfffffhh")
    def _put_tattaxis(self, cursor, axis):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        vers = 4
        buff = (cursor.put_fields(self._format_tattaxis,
                                  axis["_fNdivisions"],
                                  axis["_fAxisColor"],
                                  axis["_fLabelColor"],
                                  axis["_fLabelFont"],
                                  axis["_fLabelOffset"],
                                  axis["_fLabelSize"],
                                  axis["_fTickLength"],
                                  axis["_fTitleOffset"],
                                  axis["_fTitleSize"],
                                  axis["_fTitleColor"],
                                  axis["_fTitleFont"]))
        length = len(buff) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers) + buff

    _format_taxis_1 = struct.Struct(">idd")
    _format_taxis_2 = struct.Struct(">iiH?")
    def _put_taxis(self, cursor, axis):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        vers = 10
        buff = (self._put_tnamed(cursor, axis["_fName"], axis["_fTitle"]) +
                self._put_tattaxis(cursor, axis) +
                cursor.put_fields(self._format_taxis_1,
                                  axis["_fNbins"],
                                  axis["_fXmin"],
                                  axis["_fXmax"]) +
                self._put_tarray(cursor, axis["_fXbins"]) +
                cursor.put_fields(self._format_taxis_2,
                                  axis["_fFirst"],
                                  axis["_fLast"],
                                  axis["_fBits2"],
                                  axis["_fTimeDisplay"]) +
                cursor.put_string(axis["_fTimeFormat"]) +
                self.util.put_objany(cursor, (axis["_fLabels"], "THashList"), self.keycursor) +
                self.util.put_objany(cursor, (axis["_fModLabs"], "TList"), self.keycursor))
        length = len(buff) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers) + buff

    _format_th1_1 = struct.Struct(">i")
    _format_th1_2 = struct.Struct(">hhdddddddd")
    _format_th1_3 = struct.Struct(">iBii")
    def _put_th1(self, cursor, name):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        vers = 8
        if len(self._fields["_fBuffer"]) != 0:
            raise NotImplementedError
        buff = (self._put_tnamed(cursor, name, self._fTitle) +
                self._put_tattline(cursor) +
                self._put_tattfill(cursor) +
                self._put_tattmarker(cursor) +
                cursor.put_fields(self._format_th1_1, self._fields["_fNcells"]) +
                self._put_taxis(cursor, self._fXaxis) +
                self._put_taxis(cursor, self._fYaxis) +
                self._put_taxis(cursor, self._fZaxis) +
                cursor.put_fields(self._format_th1_2,
                                  self._fields["_fBarOffset"],
                                  self._fields["_fBarWidth"],
                                  self._fields["_fEntries"],
                                  self._fields["_fTsumw"],
                                  self._fields["_fTsumw2"],
                                  self._fields["_fTsumwx"],
                                  self._fields["_fTsumwx2"],
                                  self._fields["_fMaximum"],
                                  self._fields["_fMinimum"],
                                  self._fields["_fNormFactor"]) +
                self._put_tarray(cursor, self._fields["_fContour"]) +
                self._put_tarray(cursor, self._fields["_fSumw2"]) +
                cursor.put_string(self._fields["_fOption"]) +
                self._put_tlist(cursor, self._fields["_fFunctions"]) +
                cursor.put_fields(self._format_th1_3,
                                  self._fields["_fBufferSize"],
                                  0,  # FIXME: empty fBuffer
                                  self._fields["_fBinStatErrOpt"],
                                  self._fields["_fStatOverflows"]))
        length = len(buff) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers) + buff

    _format_th2_1 = struct.Struct(">dddd")
    def _put_th2(self, cursor, name):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        vers = 5
        buff = (self._put_th1(cursor, name) +
                cursor.put_fields(self._format_th2_1,
                                  self._fields["_fScalefactor"],
                                  self._fields["_fTsumwy"],
                                  self._fields["_fTsumwy2"],
                                  self._fields["_fTsumwxy"]))
        length = len(buff) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers) + buff

    _format_th3_1 = struct.Struct(">ddddddd")
    def _put_th3(self, cursor, name):
        vers = 6
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        buff = (self._put_th1(cursor, name) +
                self._put_tatt3d(cursor) + cursor.put_fields(self._format_th3_1,
                                                             self._fields["_fTsumwy"],
                                                             self._fields["_fTsumwy2"],
                                                             self._fields["_fTsumwxy"],
                                                             self._fields["_fTsumwz"],
                                                             self._fields["_fTsumwz2"],
                                                             self._fields["_fTsumwxz"],
                                                             self._fields["_fTsumwyz"]))
        length = len(buff) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers) + buff

    def _put_tatt3d(self, cursor):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        cnt = numpy.int64(self._format_cntvers.size - 4) | uproot.const.kByteCountMask
        vers = 1
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers)

    def _put_th1d(self, cursor, name):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        vers = 3
        buff = self._put_th1(cursor, name) + self._put_tarray(cursor, self._valuesarray)
        length = len(buff) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers) + buff

    def _put_th2d(self, cursor, name):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        vers = 4
        buff = self._put_th2(cursor, name) + self._put_tarray(cursor, self._valuesarray)
        length = len(buff) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers) + buff

    def _put_th3d(self, cursor, name):
        copy_cursor = copy(cursor)
        cursor.skip(self._format_cntvers.size)
        vers = 4
        buff = self._put_th3(cursor, name) + self._put_tarray(cursor, self._valuesarray)
        length = len(buff) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        return copy_cursor.put_fields(self._format_cntvers, cnt, vers) + buff

    _format_tprofile = struct.Struct(">idddd")
    def _write(self, context, cursor, name, compression, key, keycursor, util):
        self.util = util
        self.util.set_obj(self)
        copy_cursor = copy(cursor)
        write_cursor = copy(cursor)
        self.keycursor = keycursor
        cursor.skip(self._format_cntvers.size)
        if "TH1" in self._fClassName.decode("utf-8"):
            vers = 3
            buff = self._put_th1(cursor, name) + self._put_tarray(cursor, self._valuesarray)
        elif "TH2" in self._fClassName.decode("utf-8"):
            vers = 4
            buff = self._put_th2(cursor, name) + self._put_tarray(cursor, self._valuesarray)
        elif "TH3" in self._fClassName.decode("utf-8"):
            vers = 4
            buff = self._put_th3(cursor, name) + self._put_tarray(cursor, self._valuesarray)
        elif "TProfile" == self._fClassName.decode("utf-8"):
            vers = 7
            buff = (self._put_th1d(cursor, name) + self._put_tarray(cursor, self._fields["_fBinEntries"]) +
                    cursor.put_fields(self._format_tprofile, self._fields["_fErrorMode"], self._fields["_fYmin"],
                                      self._fields["_fYmax"], self._fields["_fTsumwy"], self._fields["_fTsumwy2"]) +
                    self._put_tarray(cursor, self._fields["_fBinSumw2"]))
        elif "TProfile2D" == self._fClassName.decode("utf-8"):
            vers = 8
            buff = (self._put_th2d(cursor, name)
                    + self._put_tarray(cursor, self._fields["_fBinEntries"]) +
                    cursor.put_fields(self._format_tprofile, self._fields["_fErrorMode"], self._fields["_fZmin"],
                                      self._fields["_fZmax"], self._fields["_fTsumwz"], self._fields["_fTsumwz2"]) +
                    self._put_tarray(cursor, self._fields["_fBinSumw2"]))
        elif "TProfile3D" == self._fClassName.decode("utf-8"):
            vers = 8
            buff = (self._put_th3d(cursor, name)
                    + self._put_tarray(cursor, self._fields["_fBinEntries"]) +
                    cursor.put_fields(self._format_tprofile, self._fields["_fErrorMode"], self._fields["_fTmin"],
                                      self._fields["_fTmax"], self._fields["_fTsumwt"], self._fields["_fTsumwt2"]) +
                    self._put_tarray(cursor, self._fields["_fBinSumw2"]))
        length = len(buff) + self._format_cntvers.size
        cnt = numpy.int64(length - 4) | uproot.const.kByteCountMask
        givenbytes = copy_cursor.put_fields(self._format_cntvers, cnt, vers) + buff
        uproot.write.compress.write(context, write_cursor, givenbytes, compression, key, keycursor)



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot/blob/master/LICENSE

import os
from collections import namedtuple

import numpy
import pytest

import awkward
import uproot

def basest(array):
    while getattr(array, "base", None) is not None:
        array = array.base
    return array

class Test(object):
    ###################################################### double32

    def test_double32(self):
        t = uproot.open("tests/samples/demo-double32.root")["T"]
        fD64 = t.array("fD64")
        fF32 = t.array("fF32")
        fI32 = t.array("fI32")
        fI30 = t.array("fI30")
        fI28 = t.array("fI28")
        ratio_fF32 = fF32 / fD64
        ratio_fI32 = fI32 / fD64
        ratio_fI30 = fI30 / fD64
        ratio_fI28 = fI28 / fD64
        assert ratio_fF32.min() > 0.9999 and ratio_fF32.max() < 1.0001
        assert ratio_fI32.min() > 0.9999 and ratio_fI32.max() < 1.0001
        assert ratio_fI30.min() > 0.9999 and ratio_fI30.max() < 1.0001
        assert ratio_fI28.min() > 0.9999 and ratio_fI28.max() < 1.0001

    ###################################################### basket

    def test_flat_basket(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["i8"]
        interpretation = branch._normalize_interpretation(None, awkward)
        entrystart, entrystop = uproot.tree._normalize_entrystartstop(branch.numentries, None, None)
        local_entrystart, local_entrystop = branch._localentries(0, entrystart, entrystop)

        one = branch._basket(0, interpretation, local_entrystart, local_entrystop, awkward, None, None)
        two = branch._basket(0, interpretation, local_entrystart, local_entrystop, awkward, None, None)
        assert numpy.array_equal(one, numpy.array([-15, -14, -13], dtype=">i8"))
        assert basest(one) is basest(two)

        three = branch.basket(0)
        assert numpy.array_equal(three, numpy.array([-15, -14, -13], dtype=">i8"))
        assert basest(one) is not basest(three)

        buf = numpy.zeros(10, dtype=numpy.float64)
        four = branch.basket(0, interpretation.toarray(buf))
        assert numpy.array_equal(four, numpy.array([-15, -14, -13], dtype=">i8"))
        assert basest(four) is buf

    def test_regular_basket(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["ai8"]
        interpretation = branch._normalize_interpretation(None, awkward)
        entrystart, entrystop = uproot.tree._normalize_entrystartstop(branch.numentries, None, None)
        local_entrystart, local_entrystop = branch._localentries(0, entrystart, entrystop)

        one = branch._basket(0, interpretation, local_entrystart, local_entrystop, awkward, None, None)
        two = branch._basket(0, interpretation, local_entrystart, local_entrystop, awkward, None, None)
        assert numpy.array_equal(one, numpy.array([[-14, -13, -12]], dtype=">i8"))
        assert basest(one) is basest(two)

        three = branch.basket(0)
        assert numpy.array_equal(three, numpy.array([[-14, -13, -12]], dtype=">i8"))
        assert basest(one) is not basest(three)

        assert branch.basket(0, interpretation.to(todims=(3,))).shape == (1, 3)
        assert branch.basket(0, interpretation.to(todims=())).shape == (3,)
        assert branch.basket(0, interpretation.to(todims=(1,))).shape == (3, 1)
        assert branch.basket(0, interpretation.to(todims=(1, 1))).shape == (3, 1, 1)
        assert branch.basket(0, interpretation.to(todims=(1, 3))).shape == (1, 1, 3)

        buf = numpy.zeros(10, dtype=numpy.float64)
        four = branch.basket(0, interpretation.toarray(buf))
        assert numpy.array_equal(four, numpy.array([-14, -13, -12], dtype=">i8"))
        assert basest(four) is buf

    def test_irregular_basket(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["Ai8"]
        interpretation = branch._normalize_interpretation(None, awkward)
        entrystart, entrystop = uproot.tree._normalize_entrystartstop(branch.numentries, None, None)
        local_entrystart, local_entrystop = branch._localentries(0, entrystart, entrystop)

        one = branch._basket(0, interpretation, local_entrystart, local_entrystop, awkward, None, None)
        two = branch._basket(0, interpretation, local_entrystart, local_entrystop, awkward, None, None)
        assert numpy.array_equal(one[0], numpy.array([], dtype=">i8"))
        assert numpy.array_equal(one[1], numpy.array([-15], dtype=">i8"))
        assert basest(one.content) is basest(two.content)

        three = branch.basket(0)
        assert numpy.array_equal(three[0], numpy.array([], dtype=">i8"))
        assert numpy.array_equal(three[1], numpy.array([-15], dtype=">i8"))

    def test_strings_basket(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["str"]
        interpretation = branch._normalize_interpretation(None, awkward)
        entrystart, entrystop = uproot.tree._normalize_entrystartstop(branch.numentries, None, None)
        local_entrystart, local_entrystop = branch._localentries(0, entrystart, entrystop)

        one = branch.basket(0, interpretation, local_entrystart, local_entrystop)
        two = branch.basket(0, interpretation, local_entrystart, local_entrystop)

        assert one.tolist() == [b"hey-0", b"hey-1", b"hey-2", b"hey-3", b"hey-4", b"hey-5"]
        assert basest(one.content.content) is not basest(two.content.content)

        three = branch.basket(0)
        assert three.tolist() == [b"hey-0", b"hey-1", b"hey-2", b"hey-3", b"hey-4", b"hey-5"]

    ###################################################### baskets

    def test_flat_baskets(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["i8"]
        expectation = [[-15, -14, -13], [-12, -11, -10], [-9, -8, -7], [-6, -5, -4], [-3, -2, -1], [0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10, 11], [12, 13, 14]]
        assert [x.tolist() for x in branch.baskets()] == expectation
        assert [x.tolist() for x in branch.iterate_baskets()] == expectation

    def test_regular_baskets(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["ai8"]
        expectation = [[[-14, -13, -12]], [[-13, -12, -11]], [[-12, -11, -10]], [[-11, -10, -9]], [[-10, -9, -8]], [[-9, -8, -7]], [[-8, -7, -6]], [[-7, -6, -5]], [[-6, -5, -4]], [[-5, -4, -3]], [[-4, -3, -2]], [[-3, -2, -1]], [[-2, -1, 0]], [[-1, 0, 1]], [[0, 1, 2]], [[1, 2, 3]], [[2, 3, 4]], [[3, 4, 5]], [[4, 5, 6]], [[5, 6, 7]], [[6, 7, 8]], [[7, 8, 9]], [[8, 9, 10]], [[9, 10, 11]], [[10, 11, 12]], [[11, 12, 13]], [[12, 13, 14]], [[13, 14, 15]], [[14, 15, 16]], [[15, 16, 17]]]
        assert [x.tolist() for x in branch.baskets()] == expectation
        assert [x.tolist() for x in branch.iterate_baskets()] == expectation

    def test_irregular_baskets(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["Ai8"]
        expectation = [[[], [-15]], [[-15, -13]], [[-15, -13, -11]], [[-15, -13, -11, -9]], [[], [-10]], [[-10, -8]], [[-10, -8, -6]], [[-10, -8, -6, -4]], [[], [-5]], [[-5, -3]], [[-5, -3, -1]], [[-5, -3, -1, 1]], [[], [0]], [[0, 2]], [[0, 2, 4]], [[0, 2, 4, 6]], [[], [5]], [[5, 7]], [[5, 7, 9]], [[5, 7, 9, 11]], [[], [10]], [[10, 12]], [[10, 12, 14]], [[10, 12, 14, 16]]]
        assert [len(y) for x in expectation for y in x] == [0, 1, 2, 3, 4] * 6
        assert [x.tolist() for x in branch.baskets()] == expectation
        assert [x.tolist() for x in branch.iterate_baskets()] == expectation

    def test_strings_baskets(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["str"]
        expectation = [[b"hey-0", b"hey-1", b"hey-2", b"hey-3", b"hey-4", b"hey-5"], [b"hey-6", b"hey-7", b"hey-8", b"hey-9", b"hey-10"], [b"hey-11", b"hey-12", b"hey-13", b"hey-14", b"hey-15"], [b"hey-16", b"hey-17", b"hey-18", b"hey-19", b"hey-20"], [b"hey-21", b"hey-22", b"hey-23", b"hey-24", b"hey-25"], [b"hey-26", b"hey-27", b"hey-28", b"hey-29"]]
        assert [x.tolist() for x in branch.baskets()] == expectation
        assert [x.tolist() for x in branch.iterate_baskets()] == expectation

    ###################################################### array

    def test_flat_array(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["i8"]
        expectation = [-15, -14, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert branch.array(entrystart=entrystart, entrystop=entrystop).tolist() == expectation[entrystart:entrystop]

    def test_regular_array(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["ai8"]
        expectation = [[-14, -13, -12], [-13, -12, -11], [-12, -11, -10], [-11, -10, -9], [-10, -9, -8], [-9, -8, -7], [-8, -7, -6], [-7, -6, -5], [-6, -5, -4], [-5, -4, -3], [-4, -3, -2], [-3, -2, -1], [-2, -1, 0], [-1, 0, 1], [0, 1, 2], [1, 2, 3], [2, 3, 4], [3, 4, 5], [4, 5, 6], [5, 6, 7], [6, 7, 8], [7, 8, 9], [8, 9, 10], [9, 10, 11], [10, 11, 12], [11, 12, 13], [12, 13, 14], [13, 14, 15], [14, 15, 16], [15, 16, 17]]
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert branch.array(entrystart=entrystart, entrystop=entrystop).tolist() == expectation[entrystart:entrystop]

    def test_irregular_array(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["Ai8"]
        expectation = [[], [-15], [-15, -13], [-15, -13, -11], [-15, -13, -11, -9], [], [-10], [-10, -8], [-10, -8, -6], [-10, -8, -6, -4], [], [-5], [-5, -3], [-5, -3, -1], [-5, -3, -1, 1], [], [0], [0, 2], [0, 2, 4], [0, 2, 4, 6], [], [5], [5, 7], [5, 7, 9], [5, 7, 9, 11], [], [10], [10, 12], [10, 12, 14], [10, 12, 14, 16]]
        assert [len(x) for x in expectation] == [0, 1, 2, 3, 4] * 6
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert branch.array(entrystart=entrystart, entrystop=entrystop).tolist() == expectation[entrystart:entrystop]

    def test_strings_array(self):
        branch = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]["str"]
        expectation = [b"hey-0", b"hey-1", b"hey-2", b"hey-3", b"hey-4", b"hey-5", b"hey-6", b"hey-7", b"hey-8", b"hey-9", b"hey-10", b"hey-11", b"hey-12", b"hey-13", b"hey-14", b"hey-15", b"hey-16", b"hey-17", b"hey-18", b"hey-19", b"hey-20", b"hey-21", b"hey-22", b"hey-23", b"hey-24", b"hey-25", b"hey-26", b"hey-27", b"hey-28", b"hey-29"]
        for entrystart, entrystop in [(None, None), (1, None), (1, 2), (1, 10), (10, 11), (10, 20), (6, 12), (6, 13)]:
            assert branch.array(entrystart=entrystart, entrystop=entrystop).tolist() == expectation[entrystart:entrystop]

    ###################################################### iterate

    def test_flat_iterate(self):
        tree = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]
        expectation = [-15, -14, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
        for n in 1000, 5, 6, 7:
            assert [x.tolist() for (x,) in tree.iterate("i8", n, outputtype=tuple)] == [expectation[x : x + n] for x in range(0, len(expectation), n)]

    def test_regular_iterate(self):
        tree = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]
        expectation = [[-14, -13, -12], [-13, -12, -11], [-12, -11, -10], [-11, -10, -9], [-10, -9, -8], [-9, -8, -7], [-8, -7, -6], [-7, -6, -5], [-6, -5, -4], [-5, -4, -3], [-4, -3, -2], [-3, -2, -1], [-2, -1, 0], [-1, 0, 1], [0, 1, 2], [1, 2, 3], [2, 3, 4], [3, 4, 5], [4, 5, 6], [5, 6, 7], [6, 7, 8], [7, 8, 9], [8, 9, 10], [9, 10, 11], [10, 11, 12], [11, 12, 13], [12, 13, 14], [13, 14, 15], [14, 15, 16], [15, 16, 17]]
        for n in 1000, 5, 6, 7:
            assert [x.tolist() for (x,) in tree.iterate("ai8", n, outputtype=tuple)] == [expectation[x : x + n] for x in range(0, len(expectation), n)]

    def test_irregular_iterate(self):
        tree = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]
        expectation = [[], [-15], [-15, -13], [-15, -13, -11], [-15, -13, -11, -9], [], [-10], [-10, -8], [-10, -8, -6], [-10, -8, -6, -4], [], [-5], [-5, -3], [-5, -3, -1], [-5, -3, -1, 1], [], [0], [0, 2], [0, 2, 4], [0, 2, 4, 6], [], [5], [5, 7], [5, 7, 9], [5, 7, 9, 11], [], [10], [10, 12], [10, 12, 14], [10, 12, 14, 16]]
        for n in 1000, 5, 6, 7:
            assert [x.tolist() for (x,) in tree.iterate("Ai8", n, outputtype=tuple)] == [expectation[x : x + n] for x in range(0, len(expectation), n)]

    def test_strings_iterate(self):
        tree = uproot.open("tests/samples/sample-6.10.05-uncompressed.root")["sample"]
        expectation = [b"hey-0", b"hey-1", b"hey-2", b"hey-3", b"hey-4", b"hey-5", b"hey-6", b"hey-7", b"hey-8", b"hey-9", b"hey-10", b"hey-11", b"hey-12", b"hey-13", b"hey-14", b"hey-15", b"hey-16", b"hey-17", b"hey-18", b"hey-19", b"hey-20", b"hey-21", b"hey-22", b"hey-23", b"hey-24", b"hey-25", b"hey-26", b"hey-27", b"hey-28", b"hey-29"]
        for n in 1000, 5, 6, 7:
            assert [x.tolist() for (x,) in tree.iterate("str", n, outputtype=tuple)] == [expectation[x : x + n] for x in range(0, len(expectation), n)]

    ###################################################### old tests

    def test_branch_array(self):
        file = uproot.open("tests/samples/simple.root")
        repr(file)

        tree = file["tree"]
        repr(tree)
        repr(tree["one"])

        assert tree["one"].array().tolist() == [1, 2, 3, 4]
        assert tree["two"].array().tolist() == numpy.array([1.1, 2.2, 3.3, 4.4], dtype=numpy.float32).tolist()
        assert tree["three"].array().tolist() == [b"uno", b"dos", b"tres", b"quatro"]

        assert tree["one"].array().tolist() == [1, 2, 3, 4]
        assert tree["two"].array().tolist() == numpy.array([1.1, 2.2, 3.3, 4.4], dtype=numpy.float32).tolist()
        assert tree["three"].array().tolist() == [b"uno", b"dos", b"tres", b"quatro"]

        tree = file["tree"]
        assert tree["one"].array().tolist() == [1, 2, 3, 4]
        assert tree["two"].array().tolist() == numpy.array([1.1, 2.2, 3.3, 4.4], dtype=numpy.float32).tolist()
        assert tree["three"].array().tolist() == [b"uno", b"dos", b"tres", b"quatro"]

    def test_tree_arrays(self):
        file = uproot.open("tests/samples/simple.root")

        tree = file["tree"]
        arrays = tree.arrays()
        assert arrays[b"one"].tolist() == [1, 2, 3, 4]
        assert arrays[b"two"].tolist() == numpy.array([1.1, 2.2, 3.3, 4.4], dtype=numpy.float32).tolist()
        assert arrays[b"three"].tolist() == [b"uno", b"dos", b"tres", b"quatro"]

        # get arrays again
        arrays = tree.arrays()
        assert arrays[b"one"].tolist() == [1, 2, 3, 4]
        assert arrays[b"two"].tolist() == numpy.array([1.1, 2.2, 3.3, 4.4], dtype=numpy.float32).tolist()
        assert arrays[b"three"].tolist() == [b"uno", b"dos", b"tres", b"quatro"]

        # get tree again
        tree = file["tree"]
        arrays = tree.arrays()
        assert arrays[b"one"].tolist() == [1, 2, 3, 4]
        assert arrays[b"two"].tolist() == numpy.array([1.1, 2.2, 3.3, 4.4], dtype=numpy.float32).tolist()
        assert arrays[b"three"].tolist() == [b"uno", b"dos", b"tres", b"quatro"]

    def test_tree_arrays_namedecode(self):
        file = uproot.open("tests/samples/simple.root")

        tree = file["tree"]
        arrays = tree.arrays(namedecode="utf-8")
        assert arrays["one"].tolist() == [1, 2, 3, 4]
        assert arrays["two"].tolist() == numpy.array([1.1, 2.2, 3.3, 4.4], dtype=numpy.float32).tolist()
        assert arrays["three"].tolist() == [b"uno", b"dos", b"tres", b"quatro"]

    def test_tree_iterator1(self):
        # one big array
        for arrays in uproot.open("tests/samples/foriter.root")["foriter"].iterate(entrysteps=1000):
            assert arrays[b"data"].tolist() == list(range(46))

        # size is equal to basket size (for most baskets)
        i = 0
        for arrays in uproot.open("tests/samples/foriter.root")["foriter"].iterate(entrysteps=6):
            assert arrays[b"data"].tolist() == list(range(i, min(i + 6, 46)))
            i += 6

        # size is smaller
        i = 0
        for arrays in uproot.open("tests/samples/foriter.root")["foriter"].iterate(entrysteps=3):
            assert arrays[b"data"].tolist() == list(range(i, min(i + 3, 46)))
            i += 3
        i = 0
        for arrays in uproot.open("tests/samples/foriter.root")["foriter"].iterate(entrysteps=4):
            assert arrays[b"data"].tolist() == list(range(i, min(i + 4, 46)))
            i += 4

        # size is larger
        i = 0
        for arrays in uproot.open("tests/samples/foriter.root")["foriter"].iterate(entrysteps=12):
            assert arrays[b"data"].tolist() == list(range(i, min(i + 12, 46)))
            i += 12
        i = 0
        for arrays in uproot.open("tests/samples/foriter.root")["foriter"].iterate(entrysteps=10):
            assert arrays[b"data"].tolist() == list(range(i, min(i + 10, 46)))
            i += 10

        # singleton case
        i = 0
        for arrays in uproot.open("tests/samples/foriter.root")["foriter"].iterate(entrysteps=1):
            assert arrays[b"data"].tolist() == list(range(i, min(i + 1, 46)))
            i += 1

    def test_tree_iterator2(self):
        words = [b"zero", b"one", b"two", b"three", b"four", b"five", b"six", b"seven", b"eight", b"nine", b"ten", b"eleven", b"twelve", b"thirteen", b"fourteen", b"fifteen", b"sixteen", b"seventeen", b"eighteen", b"ninteen", b"twenty", b"twenty-one", b"twenty-two", b"twenty-three", b"twenty-four", b"twenty-five", b"twenty-six", b"twenty-seven", b"twenty-eight", b"twenty-nine", b"thirty"]

        # one big array
        for arrays in uproot.open("tests/samples/foriter2.root")["foriter2"].iterate(entrysteps=1000):
            assert arrays[b"data"].tolist() == words

        # size is equal to basket size (for most baskets)
        i = 0
        for arrays in uproot.open("tests/samples/foriter2.root")["foriter2"].iterate(entrysteps=6):
            assert arrays[b"data"].tolist() == words[i:i + 6]
            i += 6

        # size is smaller
        i = 0
        for arrays in uproot.open("tests/samples/foriter2.root")["foriter2"].iterate(entrysteps=3):
            assert arrays[b"data"].tolist() == words[i:i + 3]
            i += 3
        i = 0
        for arrays in uproot.open("tests/samples/foriter2.root")["foriter2"].iterate(entrysteps=4):
            assert arrays[b"data"].tolist() == words[i:i + 4]
            i += 4

        # size is larger
        i = 0
        for arrays in uproot.open("tests/samples/foriter2.root")["foriter2"].iterate(entrysteps=12):
            assert arrays[b"data"].tolist() == words[i:i + 12]
            i += 12
        i = 0
        for arrays in uproot.open("tests/samples/foriter2.root")["foriter2"].iterate(entrysteps=10):
            assert arrays[b"data"].tolist() == words[i:i + 10]
            i += 10

        # singleton case
        i = 0
        for arrays in uproot.open("tests/samples/foriter2.root")["foriter2"].iterate(entrysteps=1):
            assert arrays[b"data"].tolist() == words[i:i + 1]
            i += 1

    def test_tree_iterator3(self):
        source = list(range(46))

        # one big array
        for arrays in uproot.iterate(["tests/samples/foriter.root", "tests/samples/foriter.root"], "foriter", entrysteps=1000):
            assert arrays[b"data"].tolist() == source

        # size is equal to basket size (for most baskets)
        i = 0
        for arrays in uproot.iterate(["tests/samples/foriter.root", "tests/samples/foriter.root"], "foriter", entrysteps=6):
            assert arrays[b"data"].tolist() == source[i : i + 6]
            i += 6
            if i > 45: i = 0

        # size is smaller
        i = 0
        for arrays in uproot.iterate(["tests/samples/foriter.root", "tests/samples/foriter.root"], "foriter", entrysteps=3):
            assert arrays[b"data"].tolist() == source[i : i + 3]
            i += 3
            if i > 45: i = 0
        i = 0
        for arrays in uproot.iterate(["tests/samples/foriter.root", "tests/samples/foriter.root"], "foriter", entrysteps=4):
            assert arrays[b"data"].tolist() == source[i : i + 4]
            i += 4
            if i > 45: i = 0

        # size is larger
        i = 0
        for arrays in uproot.iterate(["tests/samples/foriter.root", "tests/samples/foriter.root"], "foriter", entrysteps=12):
            assert arrays[b"data"].tolist() == source[i : i + 12]
            i += 12
            if i > 45: i = 0
        i = 0
        for arrays in uproot.iterate(["tests/samples/foriter.root", "tests/samples/foriter.root"], "foriter", entrysteps=10):
            assert arrays[b"data"].tolist() == source[i : i + 10]
            i += 10
            if i > 45: i = 0

        # singleton case
        i = 0
        for arrays in uproot.iterate(["tests/samples/foriter.root", "tests/samples/foriter.root"], "foriter", entrysteps=1):
            assert arrays[b"data"].tolist() == source[i : i + 1]
            i += 1
            if i > 45: i = 0

    def test_tree_iterator4(self):
        words2 = [b"zero", b"one", b"two", b"three", b"four", b"five", b"six", b"seven", b"eight", b"nine", b"ten", b"eleven", b"twelve", b"thirteen", b"fourteen", b"fifteen", b"sixteen", b"seventeen", b"eighteen", b"ninteen", b"twenty", b"twenty-one", b"twenty-two", b"twenty-three", b"twenty-four", b"twenty-five", b"twenty-six", b"twenty-seven", b"twenty-eight", b"twenty-nine", b"thirty"]

        # one big array
        for arrays in uproot.iterate(["tests/samples/foriter2.root", "tests/samples/foriter2.root"], "foriter2", entrysteps=1000):
            assert arrays[b"data"].tolist() == words2

        # size is equal to basket size (for most baskets)
        i = 0
        for arrays in uproot.iterate(["tests/samples/foriter2.root", "tests/samples/foriter2.root"], "foriter2", entrysteps=6):
            assert arrays[b"data"].tolist() == words2[i : i + 6]
            i += 6
            if i > 30: i = 0

        # size is smaller
        i = 0
        for arrays in uproot.iterate(["tests/samples/foriter2.root", "tests/samples/foriter2.root"], "foriter2", entrysteps=3):
            assert arrays[b"data"].tolist() == words2[i : i + 3]
            i += 3
            if i > 30: i = 0
        i = 0
        for arrays in uproot.iterate(["tests/samples/foriter2.root", "tests/samples/foriter2.root"], "foriter2", entrysteps=4):
            assert arrays[b"data"].tolist() == words2[i : i + 4]
            i += 4
            if i > 30: i = 0

        # size is larger
        i = 0
        for arrays in uproot.iterate(["tests/samples/foriter2.root", "tests/samples/foriter2.root"], "foriter2", entrysteps=12):
            assert arrays[b"data"].tolist() == words2[i : i + 12]
            i += 12
            if i > 30: i = 0
        i = 0
        for arrays in uproot.iterate(["tests/samples/foriter2.root", "tests/samples/foriter2.root"], "foriter2", entrysteps=10):
            assert arrays[b"data"].tolist() == words2[i : i + 10]
            i += 10
            if i > 30: i = 0

        # singleton case
        i = 0
        for arrays in uproot.iterate(["tests/samples/foriter2.root", "tests/samples/foriter2.root"], "foriter2", entrysteps=1):
            assert arrays[b"data"].tolist() == words2[i : i + 1]
            i += 1
            if i > 30: i = 0

    def test_directories(self):
        file = uproot.open("tests/samples/nesteddirs.root")

        assert [(n, cls._classname) for n, cls in file.classes()] == [(b"one;1", b"TDirectory"), (b"three;1", b"TDirectory")]
        assert [(n, cls._classname) for n, cls in file.allclasses()] == [(b"one;1", b"TDirectory"), (b"one/two;1", b"TDirectory"), (b"one/two/tree;1", b"TTree"), (b"one/tree;1", b"TTree"), (b"three;1", b"TDirectory"), (b"three/tree;1", b"TTree")]

        assert list(file["one"]["tree"].keys()) == [b"one", b"two", b"three"]
        assert list(file["one"].get("tree", 1).keys()) == [b"one", b"two", b"three"]
        assert list(file["one/tree;1"].keys()) == [b"one", b"two", b"three"]
        assert list(file["one/two/tree;1"].keys()) == [b"Int32", b"Int64", b"UInt32", b"UInt64", b"Float32", b"Float64", b"Str", b"ArrayInt32", b"ArrayInt64", b"ArrayUInt32", b"ArrayUInt64", b"ArrayFloat32", b"ArrayFloat64", b"N", b"SliceInt32", b"SliceInt64", b"SliceUInt32", b"SliceUInt64", b"SliceFloat32", b"SliceFloat64"]
        assert list(file["three/tree;1"].keys()) == [b"evt"]

        assert dict((name, array.tolist()) for name, array in file["one/tree"].arrays(["one", "two", "three"]).items()) == {b"one": [1, 2, 3, 4], b"two": [1.100000023841858, 2.200000047683716, 3.299999952316284, 4.400000095367432], b"three": [b"uno", b"dos", b"tres", b"quatro"]}
        assert file["one/two/tree"].array("Int32").shape == (100,)
        assert file["three/tree"].array("I32").shape == (100,)

        file = uproot.open("tests/samples/nesteddirs.root")

        assert list(file["one/tree"].keys()) == [b"one", b"two", b"three"]
        assert list(file["one/two/tree"].keys()) == [b"Int32", b"Int64", b"UInt32", b"UInt64", b"Float32", b"Float64", b"Str", b"ArrayInt32", b"ArrayInt64", b"ArrayUInt32", b"ArrayUInt64", b"ArrayFloat32", b"ArrayFloat64", b"N", b"SliceInt32", b"SliceInt64", b"SliceUInt32", b"SliceUInt64", b"SliceFloat32", b"SliceFloat64"]
        assert list(file["three/tree"].keys()) == [b"evt"]

        assert dict((name, array.tolist()) for name, array in file["one/tree;1"].arrays(["one", "two", "three"]).items()) == {b"one": [1, 2, 3, 4], b"two": [1.100000023841858, 2.200000047683716, 3.299999952316284, 4.400000095367432], b"three": [b"uno", b"dos", b"tres", b"quatro"]}
        assert file["one/two/tree;1"].array("Int32").shape == (100,)
        assert file["three/tree;1"].array("I32").shape == (100,)

    def test_cast(self):
        tree = uproot.open("tests/samples/Zmumu.root")["events"]
        one = numpy.cast[numpy.int32](numpy.floor(tree.array("M")))
        two = tree.array("M", numpy.int32)
        assert one.dtype == two.dtype
        assert one.shape == two.shape
        assert numpy.array_equal(one, two)

        for (one,) in tree.iterate("M", 10000, outputtype=tuple):
            one = numpy.cast[numpy.int32](numpy.floor(one))
        for (two,) in tree.iterate({"M": numpy.int32}, 10000, outputtype=tuple):
            pass
        assert one.dtype == two.dtype
        assert one.shape == two.shape
        assert numpy.array_equal(one, two)

    def test_pass_array(self):
        tree = uproot.open("tests/samples/Zmumu.root")["events"]
        one = numpy.cast[numpy.int32](numpy.floor(tree.array("M")))
        two = numpy.zeros(one.shape, dtype=one.dtype)
        tree.array("M", two)
        assert numpy.array_equal(one, two)

        for (one,) in tree.iterate("M", 10000, outputtype=tuple):
            one = numpy.cast[numpy.int32](numpy.floor(one))
            two = numpy.zeros(one.shape, dtype=one.dtype)
            for (two,) in tree.iterate({"M": numpy.int32}, 10000, outputtype=tuple):
                assert numpy.array_equal(one, two)

    def test_outputtype(self):
        tree = uproot.open("tests/samples/simple.root")["tree"]

        arrays = tree.arrays(["three", "two", "one"], outputtype=dict)
        assert isinstance(arrays, dict)
        assert arrays[b"one"].tolist() == [1, 2, 3, 4]
        assert arrays[b"three"].tolist() == [b"uno", b"dos", b"tres", b"quatro"]

        arrays = tree.arrays(["three", "two", "one"], outputtype=tuple)
        assert isinstance(arrays, tuple)
        assert arrays[2].tolist() == [1, 2, 3, 4]
        assert arrays[0].tolist() == [b"uno", b"dos", b"tres", b"quatro"]

        arrays = tree.arrays(["three", "two", "one"], outputtype=namedtuple)
        assert arrays.one.tolist() == [1, 2, 3, 4]
        assert arrays.three.tolist() == [b"uno", b"dos", b"tres", b"quatro"]

        arrays = tree.arrays(["three", "two", "one"], outputtype=list)
        assert isinstance(arrays, list)
        assert arrays[2].tolist() == [1, 2, 3, 4]
        assert arrays[0].tolist() == [b"uno", b"dos", b"tres", b"quatro"]

        class Awesome(object):
            def __init__(self, one, two, three):
                self.one = one
                self.two = two
                self.three = three

        arrays = tree.arrays(["one", "two", "three"], outputtype=Awesome)
        assert isinstance(arrays, Awesome)
        assert arrays.one.tolist() == [1, 2, 3, 4]
        assert arrays.three.tolist() == [b"uno", b"dos", b"tres", b"quatro"]

        class MyList(list):
            pass

        class MyTuple(tuple):
            pass

        arrays = tree.arrays(["three", "two", "one"], outputtype=MyList)
        assert isinstance(arrays, MyList)
        assert arrays[2].tolist() == [1, 2, 3, 4]
        assert arrays[0].tolist() == [b"uno", b"dos", b"tres", b"quatro"]

        arrays = tree.arrays(["three", "two", "one"], outputtype=MyTuple)
        assert isinstance(arrays, MyTuple)

    def test_tree_lazy(self):
        tree = uproot.open("tests/samples/sample-5.30.00-uncompressed.root")["sample"]

        for branchname in b"u1", b"i8", b"Ai8", b"f4", b"af4":
            strict = tree[branchname].array()

            lazy = tree[branchname].lazyarray()
            for i in range(len(lazy)):
                assert lazy[i].tolist() == strict[i].tolist()

            lazy = tree[branchname].lazyarray()
            for i in range(len(lazy), 0, -1):
                assert lazy[i - 1].tolist() == strict[i - 1].tolist()

            lazy = tree[branchname].lazyarray()
            for i in range(len(lazy)):
                assert lazy[i : i + 3].tolist() == strict[i : i + 3].tolist()

            lazy = tree[branchname].lazyarray()
            for i in range(len(lazy), 0, -1):
                assert lazy[i - 1 : i + 3].tolist() == strict[i - 1 : i + 3].tolist()

    def test_tree_lazy2(self):
        tree = uproot.open("tests/samples/sample-5.30.00-uncompressed.root")["sample"]
        lazy = tree.lazyarrays()

        for branchname in "u1", "i8", "Ai8", "f4", "af4":
            strict = tree[branchname.encode()].array()

            for i in range(len(lazy)):
                assert lazy[branchname][i].tolist() == strict[i].tolist()

            for i in range(len(lazy), 0, -1):
                assert lazy[branchname][i - 1].tolist() == strict[i - 1].tolist()

            for i in range(len(lazy)):
                assert lazy[branchname][i : i + 3].tolist() == strict[i : i + 3].tolist()

            for i in range(len(lazy), 0, -1):
                assert lazy[branchname][i - 1 : i + 3].tolist() == strict[i - 1 : i + 3].tolist()

    def test_tree_lazy3(self):
        lazy = uproot.lazyarrays(["tests/samples/sample-5.29.02-uncompressed.root", "tests/samples/sample-5.30.00-uncompressed.root"], "sample")

        assert lazy["u1"].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29]
        assert lazy["i8"].tolist() == [-15, -14, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, -15, -14, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
        assert lazy["Ai8"].tolist() == [[], [-15], [-15, -13], [-15, -13, -11], [-15, -13, -11, -9], [], [-10], [-10, -8], [-10, -8, -6], [-10, -8, -6, -4], [], [-5], [-5, -3], [-5, -3, -1], [-5, -3, -1, 1], [], [0], [0, 2], [0, 2, 4], [0, 2, 4, 6], [], [5], [5, 7], [5, 7, 9], [5, 7, 9, 11], [], [10], [10, 12], [10, 12, 14], [10, 12, 14, 16], [], [-15], [-15, -13], [-15, -13, -11], [-15, -13, -11, -9], [], [-10], [-10, -8], [-10, -8, -6], [-10, -8, -6, -4], [], [-5], [-5, -3], [-5, -3, -1], [-5, -3, -1, 1], [], [0], [0, 2], [0, 2, 4], [0, 2, 4, 6], [], [5], [5, 7], [5, 7, 9], [5, 7, 9, 11], [], [10], [10, 12], [10, 12, 14], [10, 12, 14, 16]]
        assert lazy["f4"].tolist() == [-14.899999618530273, -13.899999618530273, -12.899999618530273, -11.899999618530273, -10.899999618530273, -9.899999618530273, -8.899999618530273, -7.900000095367432, -6.900000095367432, -5.900000095367432, -4.900000095367432, -3.9000000953674316, -2.9000000953674316, -1.899999976158142, -0.8999999761581421, 0.10000000149011612, 1.100000023841858, 2.0999999046325684, 3.0999999046325684, 4.099999904632568, 5.099999904632568, 6.099999904632568, 7.099999904632568, 8.100000381469727, 9.100000381469727, 10.100000381469727, 11.100000381469727, 12.100000381469727, 13.100000381469727, 14.100000381469727, -14.899999618530273, -13.899999618530273, -12.899999618530273, -11.899999618530273, -10.899999618530273, -9.899999618530273, -8.899999618530273, -7.900000095367432, -6.900000095367432, -5.900000095367432, -4.900000095367432, -3.9000000953674316, -2.9000000953674316, -1.899999976158142, -0.8999999761581421, 0.10000000149011612, 1.100000023841858, 2.0999999046325684, 3.0999999046325684, 4.099999904632568, 5.099999904632568, 6.099999904632568, 7.099999904632568, 8.100000381469727, 9.100000381469727, 10.100000381469727, 11.100000381469727, 12.100000381469727, 13.100000381469727, 14.100000381469727]
        assert lazy["af4"].tolist() == [[-13.899999618530273, -12.899999618530273, -11.899999618530273], [-12.899999618530273, -11.899999618530273, -10.899999618530273], [-11.899999618530273, -10.899999618530273, -9.899999618530273], [-10.899999618530273, -9.899999618530273, -8.899999618530273], [-9.899999618530273, -8.899999618530273, -7.900000095367432], [-8.899999618530273, -7.900000095367432, -6.900000095367432], [-7.900000095367432, -6.900000095367432, -5.900000095367432], [-6.900000095367432, -5.900000095367432, -4.900000095367432], [-5.900000095367432, -4.900000095367432, -3.9000000953674316], [-4.900000095367432, -3.9000000953674316, -2.9000000953674316], [-3.9000000953674316, -2.9000000953674316, -1.899999976158142], [-2.9000000953674316, -1.899999976158142, -0.8999999761581421], [-1.899999976158142, -0.8999999761581421, 0.10000000149011612], [-0.8999999761581421, 0.10000000149011612, 1.100000023841858], [0.10000000149011612, 1.100000023841858, 2.0999999046325684], [1.100000023841858, 2.0999999046325684, 3.0999999046325684], [2.0999999046325684, 3.0999999046325684, 4.099999904632568], [3.0999999046325684, 4.099999904632568, 5.099999904632568], [4.099999904632568, 5.099999904632568, 6.099999904632568], [5.099999904632568, 6.099999904632568, 7.099999904632568], [6.099999904632568, 7.099999904632568, 8.100000381469727], [7.099999904632568, 8.100000381469727, 9.100000381469727], [8.100000381469727, 9.100000381469727, 10.100000381469727], [9.100000381469727, 10.100000381469727, 11.100000381469727], [10.100000381469727, 11.100000381469727, 12.100000381469727], [11.100000381469727, 12.100000381469727, 13.100000381469727], [12.100000381469727, 13.100000381469727, 14.100000381469727], [13.100000381469727, 14.100000381469727, 15.100000381469727], [14.100000381469727, 15.100000381469727, 16.100000381469727], [15.100000381469727, 16.100000381469727, 17.100000381469727], [-13.899999618530273, -12.899999618530273, -11.899999618530273], [-12.899999618530273, -11.899999618530273, -10.899999618530273], [-11.899999618530273, -10.899999618530273, -9.899999618530273], [-10.899999618530273, -9.899999618530273, -8.899999618530273], [-9.899999618530273, -8.899999618530273, -7.900000095367432], [-8.899999618530273, -7.900000095367432, -6.900000095367432], [-7.900000095367432, -6.900000095367432, -5.900000095367432], [-6.900000095367432, -5.900000095367432, -4.900000095367432], [-5.900000095367432, -4.900000095367432, -3.9000000953674316], [-4.900000095367432, -3.9000000953674316, -2.9000000953674316], [-3.9000000953674316, -2.9000000953674316, -1.899999976158142], [-2.9000000953674316, -1.899999976158142, -0.8999999761581421], [-1.899999976158142, -0.8999999761581421, 0.10000000149011612], [-0.8999999761581421, 0.10000000149011612, 1.100000023841858], [0.10000000149011612, 1.100000023841858, 2.0999999046325684], [1.100000023841858, 2.0999999046325684, 3.0999999046325684], [2.0999999046325684, 3.0999999046325684, 4.099999904632568], [3.0999999046325684, 4.099999904632568, 5.099999904632568], [4.099999904632568, 5.099999904632568, 6.099999904632568], [5.099999904632568, 6.099999904632568, 7.099999904632568], [6.099999904632568, 7.099999904632568, 8.100000381469727], [7.099999904632568, 8.100000381469727, 9.100000381469727], [8.100000381469727, 9.100000381469727, 10.100000381469727], [9.100000381469727, 10.100000381469727, 11.100000381469727], [10.100000381469727, 11.100000381469727, 12.100000381469727], [11.100000381469727, 12.100000381469727, 13.100000381469727], [12.100000381469727, 13.100000381469727, 14.100000381469727], [13.100000381469727, 14.100000381469727, 15.100000381469727], [14.100000381469727, 15.100000381469727, 16.100000381469727], [15.100000381469727, 16.100000381469727, 17.100000381469727]]

    def test_tree_lazy_cached(self):
        tree = uproot.open("tests/samples/sample-5.30.00-uncompressed.root")["sample"]

        cache = {}
        keycache = {}
        basketcache = {}

        for branchname in b"u1", b"i8", b"Ai8", b"f4", b"af4":
            strict = tree[branchname].array()

            lazy = tree[branchname].lazyarray(cache=cache, keycache=keycache, basketcache=basketcache)
            for i in range(len(lazy)):
                assert lazy[i].tolist() == strict[i].tolist()

            lazy = tree[branchname].lazyarray(cache=cache, keycache=keycache, basketcache=basketcache)
            for i in range(len(lazy), 0, -1):
                assert lazy[i - 1].tolist() == strict[i - 1].tolist()

            lazy = tree[branchname].lazyarray(cache=cache, keycache=keycache, basketcache=basketcache)
            for i in range(len(lazy)):
                assert lazy[i : i + 3].tolist() == strict[i : i + 3].tolist()

            lazy = tree[branchname].lazyarray(cache=cache, keycache=keycache, basketcache=basketcache)
            for i in range(len(lazy), 0, -1):
                assert lazy[i - 1 : i + 3].tolist() == strict[i - 1 : i + 3].tolist()

    @pytest.mark.parametrize("use_http", [False, True])
    def test_hist_in_tree(self, use_http):
        if use_http:
            pytest.importorskip("requests")
            tree = uproot.open("http://scikit-hep.org/uproot/examples/Event.root")["T"]
        else:
            path = os.path.join("tests", "samples", "Event.root")
            if not os.path.exists(path):
                raise pytest.skip()
            tree = uproot.open(path)["T"]
        check = [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0,
                 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 1.0,
                 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 1.0, 1.0,
                 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0, 1.0,
                 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0,
                 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0]

        assert tree.array("fH")[20].values.tolist() == check

    @pytest.mark.parametrize("use_http", [False, True])
    def test_branch_auto_interpretation(self, use_http):
        # The aim is to reduce this list in a controlled manner
        known_branches_without_interp = [
            b'event',
            b'TObject',
            b'fClosestDistance',
            b'fEvtHdr',
            b'fTracks',
            b'fTracks.fPointValue',
            b'fTriggerBits',
            b'fTriggerBits.TObject'
        ]
        if use_http:
            pytest.importorskip("requests")
            tree = uproot.open("http://scikit-hep.org/uproot/examples/Event.root")["T"]
        else:
            path = os.path.join("tests", "samples", "Event.root")
            if not os.path.exists(path):
                raise pytest.skip()
            tree = uproot.open(path)["T"]
        branches_without_interp = [b.name for b in tree.allvalues() if b.interpretation is None]
        assert branches_without_interp == known_branches_without_interp
        assert tree.array("fTracks.fTArray[3]", entrystop=10)[5][10].tolist()  == [11.03951644897461, 19.40645980834961, 34.54059982299805]
        assert tree.array("fTracks.fCharge", entrystop=10)[0][0:10].tolist()   == [1.0, 1.0, 1.0, 1.0,-1.0, 0.0, 1.0, 0.0, 0.0, 0.0]
        assert tree.array("fMatrix[4][4]", entrystop=10)[0][1].tolist()        == [-0.13630907237529755, 0.8007842898368835, 1.706235647201538, 0.0]
        assert tree.array("fTracks.fMass2", entrystop=10)[3][330:333].tolist() == [8.90625, 8.90625, 8.90625]
        assert tree.array("fTracks.fBx", entrystop=10)[9][10:13].tolist()      == [0.12298583984375, -0.2489013671875, -0.189697265625]
        assert tree.array("fTracks.fBy", entrystop=10)[9][10:13].tolist()      == [0.1998291015625, -0.0301513671875, 0.0736083984375]
        assert tree.array("fTracks.fXfirst", entrystop=10)[1][11:16].tolist()  == [-8.650390625, -2.8203125, -1.949951171875, 0.4859619140625, 3.0146484375]
        assert tree.array("fTracks.fXlast", entrystop=10)[1][11:16].tolist()   == [-2.18994140625, -2.64697265625, -8.4375, 1.594970703125, 6.40234375]
        assert tree.array("fTracks.fYfirst", entrystop=10)[2][22:26].tolist()  == [4.9921875, 8.46875, 1.679443359375, -6.927734375]
        assert tree.array("fTracks.fYlast", entrystop=10)[2][22:26].tolist()   == [-5.76171875, 13.7109375, 2.98583984375, -9.466796875]
        assert tree.array("fTracks.fZfirst", entrystop=10)[3][33:36].tolist()  == [53.84375, 52.3125, 48.296875]
        assert tree.array("fTracks.fZlast", entrystop=10)[3][33:36].tolist()   == [193.96875, 208.25, 228.40625]
        assert tree.array("fTracks.fVertex[3]", entrystop=10)[1][2].tolist()   == [0.245361328125, 0.029296875,-16.171875]

    def test_leaflist(self):
        tree = uproot.open("tests/samples/leaflist.root")["tree"]
        a = tree.array("leaflist")
        assert a["x"].tolist() == [1.1, 2.2, 3.3, 4.0, 5.5]   # yeah, I goofed up when making it
        assert a["y"].tolist() == [1, 2, 3, 4, 5]
        assert a["z"].tolist() == [ord("a"), ord("b"), ord("c"), ord("d"), ord("e")]

        pytest.importorskip("pandas")
        assert tree.pandas.df()["leaflist.x"].tolist() == [1.1, 2.2, 3.3, 4.0, 5.5]

        tree = uproot.open("tests/samples/HZZ-objects.root")["events"]
        tree.pandas.df("muonp4")
        tree.pandas.df("muonp4", flatten=False)
        df = tree.pandas.df("eventweight", entrystart=100, entrystop=200)
        index = df.index.tolist()
        assert min(index) == 100
        assert max(index) == 199
        df = tree.pandas.df("muonp4", entrystart=100, entrystop=200)
        index = df.index.get_level_values("entry").tolist()
        assert min(index) == 100
        assert max(index) == 199

    def test_mempartitions(self):
        t = uproot.open("tests/samples/sample-5.23.02-zlib.root")["sample"]
        assert list(t.mempartitions(500)) == [(0, 2), (2, 4), (4, 6), (6, 8), (8, 10), (10, 12), (12, 14), (14, 16), (16, 18), (18, 20), (20, 22), (22, 24), (24, 26), (26, 28), (28, 30)]
        assert [sum(y.nbytes for y in x.values()) for x in t.iterate(entrysteps="0.5 kB")] == [693, 865, 822, 779, 951, 695, 867, 824, 781, 953, 695, 867, 824, 781, 953]



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot/blob/master/LICENSE

from __future__ import absolute_import

import math
import threading
try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping

import cachetools

class ArrayCache(MutableMapping):
    @staticmethod
    def getsizeof(obj):
        return getattr(obj, "nbytes", 1)

    def __init__(self, limitbytes, method="LRU"):
        from uproot.rootio import _memsize
        m = _memsize(limitbytes)
        if m is not None:
            limitbytes = int(math.ceil(m))
        if method == "LRU":
            self._cache = cachetools.LRUCache(limitbytes, getsizeof=self.getsizeof)
        elif method == "LFU":
            self._cache = cachetools.LFUCache(limitbytes, getsizeof=self.getsizeof)
        else:
            raise ValueError("unrecognized method: {0}".format(method))

    def __contains__(self, where):
        return where in self._cache

    def __getitem__(self, where):
        return self._cache[where]

    def __setitem__(self, where, what):
        self._cache[where] = what

    def __delitem__(self, where):
        del self._cache[where]

    def __iter__(self):
        for x in self._cache:
            yield x

    def __len__(self):
        return len(self._cache)

class ThreadSafeArrayCache(ArrayCache):
    def __init__(self, limitbytes, method="LRU"):
        super(ThreadSafeArrayCache, self).__init__(limitbytes, method=method)
        self._lock = threading.Lock()

    def __contains__(self, where):
        with self._lock:
            return where in self._cache

    def __getitem__(self, where):
        with self._lock:
            return self._cache[where]

    def __setitem__(self, where, what):
        with self._lock:
            self._cache[where] = what

    def __delitem__(self, where):
        with self._lock:
            del self._cache[where]

    def __iter__(self):
        with self._lock:
            for x in self._cache:
                yield x

    def __len__(self):
        with self._lock:
            return len(self._cache)



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot/blob/master/LICENSE

from __future__ import absolute_import

import itertools
import functools
import operator

import numpy

import awkward as awkwardbase

import uproot.tree
import uproot.interp.numerical
from uproot.interp.jagged import asjagged
from uproot.interp.numerical import asdtype
from uproot.interp.objects import asobj
from uproot.interp.objects import astable

from uproot.source.memmap import MemmapSource
from uproot.source.xrootd import XRootDSource
from uproot.source.http import HTTPSource

class TTreeMethods_pandas(object):
    def __init__(self, tree):
        self._tree = tree

    def df(self, branches=None, namedecode="utf-8", entrystart=None, entrystop=None, flatten=True, flatname=None, awkwardlib=None, cache=None, basketcache=None, keycache=None, executor=None, blocking=True):
        import pandas
        return self._tree.arrays(branches=branches, outputtype=pandas.DataFrame, namedecode=namedecode, entrystart=entrystart, entrystop=entrystop, flatten=flatten, flatname=flatname, awkwardlib=awkwardlib, cache=cache, basketcache=basketcache, keycache=keycache, executor=executor, blocking=blocking)

    def iterate(self, branches=None, entrysteps=None, namedecode="utf-8", entrystart=None, entrystop=None, flatten=True, flatname=None, awkwardlib=None, cache=None, basketcache=None, keycache=None, executor=None, blocking=True):
        import pandas
        return self._tree.iterate(branches=branches, entrysteps=entrysteps, outputtype=pandas.DataFrame, namedecode=namedecode, reportentries=False, entrystart=entrystart, entrystop=entrystop, flatten=flatten, flatname=flatname, awkwardlib=awkwardlib, cache=cache, basketcache=basketcache, keycache=keycache, executor=executor, blocking=blocking)

def default_flatname(branchname, fieldname, index):
    out = branchname
    if not isinstance(branchname, str):
        out = branchname.decode("utf-8")
    if fieldname is not None:
        out += "." + fieldname
    if index != ():
        out += "[" + "][".join(str(x) for x in index) + "]"
    return out

def futures2df(futures, outputtype, entrystart, entrystop, flatten, flatname, awkward):
    import pandas

    if flatname is None:
        flatname = default_flatname

    if not flatten or all(interpretation.__class__ is not asjagged for name, interpretation, future in futures):
        columns = []
        data = {}
        for name, interpretation, future in futures:
            array = future()

            if isinstance(interpretation, asobj) and isinstance(interpretation.content, astable):
                interpretation = interpretation.content
            if isinstance(interpretation, astable) and isinstance(interpretation.content, asdtype):
                interpretation = interpretation.content

            if isinstance(interpretation, asdtype):
                if interpretation.todims == ():
                    if interpretation.todtype.names is None:
                        fn = flatname(name, None, ())
                        columns.append(fn)
                        data[fn] = array
                    else:
                        for nn in interpretation.todtype.names:
                            if not nn.startswith(" "):
                                fn = flatname(name, nn, ())
                                columns.append(fn)
                                data[fn] = array[nn]
                else:
                    for tup in itertools.product(*[range(x) for x in interpretation.todims]):
                        if interpretation.todtype.names is None:
                            fn = flatname(name, None, tup)
                            columns.append(fn)
                            data[fn] = array[(slice(None),) + tup]
                        else:
                            for nn in interpretation.todtype.names:
                                if not nn.startswith(" "):
                                    fn = flatname(name, nn, tup)
                                    columns.append(fn)
                                    data[fn] = array[nn][(slice(None),) + tup]
            else:
                fn = flatname(name, None, ())
                columns.append(fn)
                data[fn] = list(array)     # must be serialized as a Python list for Pandas to accept it

        index = pandas.RangeIndex(entrystart, entrystop, name="entry")
        return outputtype(columns=columns, data=data, index=index)

    else:
        starts, stops = None, None

        needbroadcasts = []
        names = []
        interpretations = []
        arrays = []
        for name, interpretation, future in futures:
            if isinstance(interpretation, asobj) and isinstance(interpretation.content, astable):
                interpretation = interpretation.content
            if isinstance(interpretation, astable) and isinstance(interpretation.content, asdtype):
                interpretation = interpretation.content

            array = future()
            if isinstance(interpretation, asjagged):
                interpretation = interpretation.content
                if isinstance(interpretation, asobj) and isinstance(interpretation.content, astable):
                    interpretation = interpretation.content
                if isinstance(interpretation, astable) and isinstance(interpretation.content, asdtype):
                    interpretation = interpretation.content

                # justifies the assumption that array.content == array.flatten() and array.stops.max() == array.stops[-1]
                assert len(array.starts) == 0 or ((array.offsetsaliased(array.starts, array.stops) or (len(array.starts.shape) == 1 and array.numpy.array_equal(array.starts[1:], array.stops[:-1]))) and array.starts[0] == 0)

                if starts is None:
                    starts = array.starts
                    stops = array.stops
                    index = array.localindex
                else:
                    if starts is not array.starts and not awkward.numpy.array_equal(starts, array.starts):
                        raise ValueError("cannot use flatten=True on branches with different jagged structure, such as electrons and muons (different, variable number of each per event); either explicitly select compatible branches, such as [\"MET_*\", \"Muon_*\"] (scalar and variable per event is okay), or set flatten=False")

                if len(array.starts) == 0:
                    array = array.content[0:0]
                else:
                    array = array.content
                needbroadcasts.append(False)

            else:
                needbroadcasts.append(True)

            names.append(name)
            interpretations.append(interpretation)
            arrays.append(array)

        index = pandas.MultiIndex.from_arrays([index.tojagged(numpy.arange(entrystart, entrystop, dtype=numpy.int64)).content, index.content], names=["entry", "subentry"])

        df = outputtype(index=index)

        for name, interpretation, array, needbroadcast in zip(names, interpretations, arrays, needbroadcasts):
            if isinstance(interpretation, uproot.interp.numerical._asnumeric):
                if isinstance(array, awkwardbase.ObjectArray):
                    array = array.content

                if needbroadcast:
                    # Invoke jagged broadcasting to align arrays
                    originaldtype = array.dtype
                    originaldims = array.shape[1:]

                    if isinstance(array, awkwardbase.Table):
                        for nn in array.columns:
                            array[nn] = awkward.JaggedArray(starts, stops, awkward.numpy.empty(stops[-1], dtype=array[nn].dtype)).tojagged(array[nn]).content

                    else:
                        if len(originaldims) != 0:
                            array = array.view(awkward.numpy.dtype([(str(i), array.dtype) for i in range(functools.reduce(operator.mul, array.shape[1:]))])).reshape(array.shape[0])

                        array = awkward.JaggedArray(starts, stops, awkward.numpy.empty(stops[-1], dtype=array.dtype)).tojagged(array).content
                        if len(originaldims) != 0:
                            array = array.view(originaldtype).reshape((-1,) + originaldims)

                if interpretation.todims == ():
                    if interpretation.todtype.names is None:
                        fn = flatname(name, None, ())
                        df[fn] = array
                    else:
                        for nn in interpretation.todtype.names:
                            if not nn.startswith(" "):
                                fn = flatname(name, nn, ())
                                df[fn] = array[nn]
                else:
                    for tup in itertools.product(*[range(x) for x in interpretation.todims]):
                        if interpretation.todtype.names is None:
                            fn = flatname(name, None, tup)
                            df[fn] = array[(slice(None),) + tup]
                        else:
                            for nn in interpretation.todtype.names:
                                if not nn.startswith(" "):
                                    fn = flatname(name, nn, tup)
                                    df[fn] = array[nn][(slice(None),) + tup]

            else:
                fn = flatname(name, None, ())

                array = awkward.numpy.array(array, dtype=object)
                indexes = numpy.arange(len(array))

                indexes = awkward.JaggedArray(starts, stops, awkward.numpy.empty(stops[-1], dtype=object)).tojagged(indexes).content

                array = array[indexes]

                if len(array) != 0 and isinstance(array[0], awkward.numpy.ndarray):
                    df[fn] = list(array)
                else:
                    df[fn] = array

        return df
